"""
Struct Module.

Contains flexible objects that can hold different properties.

Used to avoid creating specific classes for each possible type.
"""

from typing import Any, List, Optional, Dict, Callable
from xml.etree import ElementTree as ET
from pineboolib.core.utils import utils_base


class Struct(object):
    """
    Basic object template.

    Sets its properties in __init__.
    Specially useful to sketch classes on the fly.
    """

    fields: List[str]
    pk: List[str]
    fields_idx: Dict[str, int]

    def __init__(self, **kwargs: Any) -> None:
        """Construct a new object using the arguments provided."""
        for key, value in kwargs.items():
            setattr(self, key, value)


class XMLStruct(Struct):
    """
    Object template replicating xml contents.

    Useful to quickly build objects that are like the parsed xml.

    Can be easily accessed by properties.
    """

    def __init__(self, xmlobj: Optional["ET.Element"] = None) -> None:
        """
        Build instance from parsed xml.
        """
        self._attrs: List[str] = []
        if xmlobj is not None:
            self.__name__ = xmlobj.tag
            for child in xmlobj:
                if child.tag == "property":
                    # Se importa aquí para evitar error de importación cíclica.
                    raise Exception("FIXME: No property support")
                    # FIXME: Esto es del DGI QT:
                    # from pineboolib.pnqt3ui import loadProperty
                    # key, text = loadProperty(child)
                else:
                    text = utils_base.AQTT(child.text)
                    key = child.tag
                if isinstance(text, str):
                    text = text.strip()
                try:
                    setattr(self, key, text)
                    self._attrs.append(key)
                except Exception:
                    print("utils.XMLStruct: Omitiendo", self.__name__, key, text)

    def __str__(self) -> str:
        """Create string representation."""
        attrs = ["%s=%s" % (k, repr(getattr(self, k))) for k in self._attrs]
        txtattrs = " ".join(attrs)
        return "<%s.%s %s>" % (self.__class__.__name__, self.__name__, txtattrs)

    def _v(self, k: str, default: None = None) -> Optional[str]:
        """Return optional value with sensible default."""
        return getattr(self, k, default)

    def _rv(self, k: str) -> str:
        """Return required key and throw error if does not exist or is not a string."""
        ret = getattr(self, k, None)
        if not isinstance(ret, str):
            raise ValueError("Retrieving value for %s, found %r which is not a string" % (k, ret))
        return ret


class AreaStruct(Struct):
    """Struct version for Module Areas."""

    idarea: str
    descripcion: str


class TableStruct(Struct):
    """Struct version for Tables."""

    xmltree: Any
    xmlroot: Any
    tablename: str
    name: str
    query_table: Optional[str]
    fields: List[str]
    pk: List[str]
    fields_idx: Dict[str, int]


class ActionStruct(XMLStruct):
    """
    Struct version for Actions.
    """

    _alias: str = ""
    _name: str = ""
    _table: str = ""
    _description: str = ""
    _caption: str = ""
    _master_form: str = ""
    _master_script: str = ""
    _record_form: str = ""
    _record_script: str = ""
    _class_script: str = ""
    _class_orm: str = ""

    load: "Callable"
    load_record: "Callable"
    openDefaultForm: "Callable"
    openDefaultFormRecord: "Callable"
    formRecordWidget: "Callable"
    execMainScript: "Callable"
    execDefaultScript: "Callable"
    unknownSlot: "Callable"
    setCursor: "Callable"
    cursor: "Callable"



"""
Test QS Snippets.
"""
from PyQt6 import QtCore, QtWidgets  # type: ignore[import]

import unittest
from pineboolib.application.parsers.parser_qsa.postparse import pythonify_string as qs2py
from pineboolib.application.parsers import parser_qsa
from pineboolib.application.parsers.parser_qsa.tests import fixture_read, fixture_path
from pineboolib.loader.main import init_testing, finish_testing

from pineboolib.core.utils import utils_base
import os
import shutil


class TestParser(unittest.TestCase):
    """Test Parsing QS to PY."""

    @classmethod
    def setUpClass(cls) -> None:
        """Enable strict parsing."""

        parser_qsa.STRICT_MODE = True
        init_testing()

    def test_basic(self) -> None:
        """Test basic stuff."""

        self.assertEqual(qs2py("x = 0"), "x = 0\n")

    def test_aqssproject(self) -> None:
        """Test aqssproject parse."""

        self.assertEqual(qs2py("QSProject.entryFunction"), "qsa.QSProject.entryFunction\n")

    def test_file_class(self) -> None:
        """Test parsing the file class."""
        self.assertEqual(qs2py('x = File.read("test")'), 'x = qsa.FileStatic.read("test")\n')
        self.assertEqual(
            qs2py('x = File.write("test", "contents")'),
            'x = qsa.FileStatic.write("test", "contents")\n',
        )
        self.assertEqual(qs2py('x = File.remove("test")'), 'x = qsa.FileStatic.remove("test")\n')

        self.assertEqual(qs2py('x = File("test").read()'), 'x = qsa.File("test").read()\n')
        self.assertEqual(
            qs2py('x = File("test").write("contents")'), 'x = qsa.File("test").write("contents")\n'
        )
        self.assertEqual(qs2py('x = File("test").remove()'), 'x = qsa.File("test").remove()\n')

    def test_list_arrays(self) -> None:
        """Test parsing iterable classes."""
        self.assertEqual(qs2py("var value = Array().shift()"), "value: Any = qsa.Array().pop(0)\n")
        self.assertEqual(qs2py("var value = [].shift()"), "value: Any = [].pop(0)\n")

    def test_array(self) -> None:
        """Test array."""

        self.assertEqual(qs2py("var a = new Array();"), "a: Any = qsa.Array()\n")
        self.assertEqual(qs2py("var b = new Array(0);"), "b: Any = []\n")

    def test_typeof(self) -> None:
        """Test typeof."""

        self.assertEqual(qs2py("var a = typeof('juan')"), 'a: Any = qsa.typeof_("juan")\n')
        self.assertEqual(
            qs2py('debug("hola " + " " + typeof("juan"));'),
            'qsa.debug(qsa.ustr("hola ", " ", qsa.typeof_("juan")))\n',
        )

    def test_process_class(self) -> None:
        """Test parsing the process class."""
        self.assertEqual(
            qs2py('x = Process.execute(["ls", "*"])'),
            'x = qsa.ProcessStatic.execute(qsa.Array(["ls", "*"]))\n',
        )

    def test_while(self) -> None:
        """Test while class."""
        value = "with (this.iface.curFactura)"
        value += (
            'setValueBuffer("neto", formfacturascli.iface.pub_commonCalculateField("neto", this));'
        )
        result_value = "# WITH_START\n"
        result_value += "self.iface.curFactura.setValueBuffer(\n"
        result_value += '    "neto", qsa.from_project("formfacturascli").iface.pub_commonCalculateField("neto", self.iface.curFactura)\n'
        result_value += ")\n"
        result_value += "# WITH_END\n"

        self.assertEqual(qs2py(value), result_value)

    def test_flfacturac(self) -> None:
        """Test conveting fixture flfacturac."""

        flfacturac_qs = fixture_read("flfacturac.qs")
        flfacturac_py = fixture_read("flfacturac.python")
        flfacturac_qs_py = qs2py(flfacturac_qs, parser_template="file_template")

        # Delete version translator tag.
        pos_ini = flfacturac_qs_py.find("# Translated with pineboolib ")
        pos_fin = flfacturac_qs_py[pos_ini:].find("\n")
        flfacturac_qs_py = flfacturac_qs_py.replace(
            flfacturac_qs_py[pos_ini : pos_ini + pos_fin + 1], ""
        )

        # Write onto git so we have an example.
        with open(fixture_path("flfacturac.qs.python"), "w") as file_:
            file_.write(flfacturac_qs_py)

        self.assertEqual(flfacturac_qs_py, flfacturac_py)

    def test_lib_str(self) -> None:
        """Test conveting fixture lib_str."""

        parser_qsa.STRICT_MODE = False
        self.maxDiff = None  # pylint: disable=invalid-name
        flfacturac_qs = fixture_read("lib_str.qs")
        flfacturac_py = fixture_read("lib_str.python")
        flfacturac_qs_py = qs2py(flfacturac_qs, parser_template="file_template")

        # Delete version translator tag.
        pos_ini = flfacturac_qs_py.find("# Translated with pineboolib ")
        pos_fin = flfacturac_qs_py[pos_ini:].find("\n")
        flfacturac_qs_py = flfacturac_qs_py.replace(
            flfacturac_qs_py[pos_ini : pos_ini + pos_fin + 1], ""
        )
        parser_qsa.STRICT_MODE = True
        # Write onto git so we have an example.
        with open(fixture_path("lib_str.qs.python"), "w") as file_:
            file_.write(flfacturac_qs_py)

        self.assertEqual(flfacturac_qs_py, flfacturac_py)

    def test_form(self) -> None:
        """Test converting form"""
        self.assertEqual(qs2py("form = this;"), "form = self\n")

    def test_parse_int(self) -> None:
        """Test parseInt function."""
        self.assertEqual(qs2py('var value = parseInt("21");'), 'value: Any = qsa.parseInt("21")\n')
        self.assertEqual(
            qs2py("var value = parseInt(2000.21 , 10);"), "value: Any = qsa.parseInt(2000.21, 10)\n"
        )
        self.assertEqual(
            qs2py('var value = parseInt("0xA0", 16);'), 'value: Any = qsa.parseInt("0xA0", 16)\n'
        )

    def test_qdir(self) -> None:
        """Test QDir translation."""
        self.assertEqual(
            qs2py(
                'var rutaImp:String = "."; var impDir = new QDir(rutaImp, "c*.csv C*.csv c*.CSV C*.CSV");'
            ),
            'rutaImp: str = "."\nimpDir: Any = qsa.QDir(rutaImp, "c*.csv C*.csv c*.CSV C*.CSV")\n',
        )

    def test_qobject(self) -> None:
        """Test QObject translation."""

        self.assertEqual(qs2py("var prueba = new QObject;"), "prueba: Any = qsa.QObject()\n")

    def test_inicialize_float(self) -> None:
        """Test float inicialization."""

        self.assertEqual(qs2py("var num:Number = 0.0;"), "num = 0.0\n")

    def test_aqsignalmapper(self) -> None:
        """Test AQSignalmapper."""

        self.assertEqual(
            qs2py("var sigMap = new AQSignalMapper(this);"),
            "sigMap: Any = qsa.AQSignalMapper(self)\n",
        )

    def test_replace(self) -> None:
        """Test replace."""

        self.assertEqual(
            qs2py(
                'var listaOutlet:Array = new Array();flfactppal.iface.replace(listaOutlet, ", ", " "," ");'
            ),
            """listaOutlet: \"qsa.Array\" = qsa.Array()
qsa.from_project("flfactppal").iface.replace(listaOutlet, ", ", " ", " ")\n""",
        )

        self.assertEqual(
            qs2py(
                "function pub_replace(cadena:String, searchValue:Number, newValue:Array)"
                + " {\nreturn this.replace(cadena, searchValue, newValue);\n}"
            ),
            """def pub_replace(self, cadena: "str", searchValue, newValue: "qsa.Array"):
    return self.replace(cadena, searchValue, newValue)\n""",
        )

    def test_optional(self) -> None:
        """Test optional."""

        self.assertEqual(
            qs2py(
                "function pub_replace(cadena:String, searchValue:Number, newValue:optional)"
                + " {\nreturn this.replace(cadena, searchValue, newValue);\n}"
            ),
            """def pub_replace(self, cadena: "str", searchValue, newValue: "Any" = None):
    return self.replace(cadena, searchValue, newValue)\n""",
        )

    def test_sort_1(self) -> None:
        """Test replace."""

        self.assertEqual(
            qs2py("var listaOutlet:Array = new Array();listaOutlet.sort();"),
            'listaOutlet: "qsa.Array" = qsa.Array()\nqsa.Sort().sort_(listaOutlet)\n',
        )

    def test_sort_2(self) -> None:
        """Test replace."""

        self.assertEqual(
            qs2py(
                """
                var aLista:Array = new Array()
                aLista.sort(parseString);
                """
            ),
            'aLista: "qsa.Array" = qsa.Array()\nqsa.Sort(qsa.parseString).sort_(aLista)\n',
        )

    def test_form2(self) -> None:
        """Test replace."""

        """Test conveting fixture flfacturac_2."""
        self.maxDiff = None
        flfacturac_qs = fixture_read("flfacturac_2.qs")
        flfacturac_py = fixture_read("flfacturac_2.python")
        flfacturac_qs_py = qs2py(flfacturac_qs, parser_template="file_template")

        # Delete version translator tag.
        pos_ini = flfacturac_qs_py.find("# Translated with pineboolib ")
        pos_fin = flfacturac_qs_py[pos_ini:].find("\n")
        flfacturac_qs_py = flfacturac_qs_py.replace(
            flfacturac_qs_py[pos_ini : pos_ini + pos_fin + 1], ""
        )

        # Write onto git so we have an example.
        with open(fixture_path("flfacturac_2.qs.python"), "w") as file_:
            file_.write(flfacturac_qs_py)

        self.assertEqual(flfacturac_qs_py, flfacturac_py)

    def test_splice(self) -> None:
        """Test splice."""

        self.assertEqual(
            qs2py(
                """
                var aLista = new Array();
                aLista.splice(10,1);
                """
            ),
            "aLista: Any = qsa.Array()\nqsa.splice(aLista, 10, 1)\n",
        )

    def test_all_any(self) -> None:
        """Test variable parse."""

        self.assertEqual(qs2py('var prueba = "hola";'), 'prueba: Any = "hola"\n')

    def test_pyconvert(self) -> None:
        """Test pyconvert."""
        from pineboolib import application

        path = fixture_path("flfacturac.qs")
        tmp_path = "%s/%s" % (application.PROJECT.tmpdir, "temp_qs.qs")
        path_py = "%s.py" % tmp_path[:-3]

        shutil.copy(path, tmp_path)
        application.PROJECT.parse_script_list([tmp_path])

        self.assertTrue(os.path.exists(path_py))

        if os.path.exists(tmp_path):
            os.remove(tmp_path)

        if os.path.exists(path_py):
            os.remove(path_py)

    def test_pyconvert_multi(self) -> None:
        """Test pyconvert simultaneously."""

        timer = QtCore.QTimer
        timer.singleShot(0, self.convert)
        timer.singleShot(0, self.convert)
        QtWidgets.QApplication.processEvents()

    def convert(self) -> None:
        """Convert a file."""

        from pineboolib import application

        path = fixture_path("flfacturac.qs")
        tmp_path = "%s/%s" % (application.PROJECT.tmpdir, "multi.qs")
        if not os.path.exists(tmp_path):
            shutil.copy(path, tmp_path)

        application.PROJECT.parse_script_list([tmp_path])

    def test_pyconvert_2(self) -> None:
        """Test conveting fixture lib_str."""

        from pineboolib import application

        self.maxDiff = None
        simple_qs_path = fixture_path("simple.qs")
        tmp_path = "%s/%s" % (application.PROJECT.tmpdir, "simple.qs")
        if not os.path.exists(tmp_path):
            shutil.copy(simple_qs_path, tmp_path)

        print("Convirtiendo", tmp_path, "existe", os.path.exists(tmp_path))
        self.assertTrue(application.PROJECT.parse_script_list([tmp_path]))
        simple_py = fixture_read("simple.python")
        simple_qs_py_path = "%spy" % tmp_path[:-2]
        self.assertTrue(os.path.exists(simple_qs_py_path))
        file_ = open(simple_qs_py_path, "r", encoding="utf-8")
        simple_qs_py = file_.read()
        file_.close()

        # Delete version translator tag.
        pos_ini = simple_qs_py.find("# Translated with pineboolib ")
        pos_fin = simple_qs_py[pos_ini:].find("\n")
        simple_qs_py = simple_qs_py.replace(simple_qs_py[pos_ini : pos_ini + pos_fin + 1], "")

        # Write onto git so we have an example.
        with open(fixture_path("simple.qs.python"), "w") as file_:
            file_.write(simple_qs_py)

        self.assertEqual(simple_qs_py, simple_py)

    def test_ignore_no_python_flag_disabled(self) -> None:
        """Test no python flag."""
        from pineboolib import application

        utils_base.FORCE_DESKTOP = False

        print(
            "Is library",
            utils_base.is_library(),
            "ignore no_python tags",
            parser_qsa.IGNORE_NO_PYTHON_TAGS,
        )
        qs_path = fixture_path("no_python.qs")
        tmp_path = "%s/%s" % (application.PROJECT.tmpdir, "no_python.qs")
        if os.path.exists(tmp_path):
            os.remove(tmp_path)
        shutil.copy(qs_path, tmp_path)

        self.assertTrue(application.PROJECT.parse_script_list([tmp_path]))
        qs_py_path = "%spy" % tmp_path[:-2]

        utils_base.FORCE_DESKTOP = True

        file_ = open(qs_py_path, "r", encoding="utf-8")
        qs_py = file_.read()
        file_.close()
        self.assertTrue(qs_py.find("TYPE_INT_") > -1)
        self.assertTrue(qs_py.find("TYPE_UINT_") == -1)

    def test_ignore_no_python_flag_enabled(self) -> None:
        """Test no python flag."""
        from pineboolib import application

        utils_base.FORCE_DESKTOP = False
        parser_qsa.IGNORE_NO_PYTHON_TAGS = True

        qs_path = fixture_path("no_python.qs")
        tmp_path = "%s/%s" % (application.PROJECT.tmpdir, "no_python2.qs")
        if os.path.exists(tmp_path):
            os.remove(tmp_path)
        shutil.copy(qs_path, tmp_path)

        self.assertTrue(application.PROJECT.parse_script_list([tmp_path]))
        qs_py_path = "%spy" % tmp_path[:-2]

        parser_qsa.IGNORE_NO_PYTHON_TAGS = False
        utils_base.FORCE_DESKTOP = True

        file_ = open(qs_py_path, "r", encoding="utf-8")
        qs_py = file_.read()
        file_.close()
        self.assertTrue(qs_py.find("TYPE_INT_") > -1)
        self.assertTrue(qs_py.find("TYPE_UINT_") > -1)

    def test_anon_class(self) -> None:
        """Test converting anon class"""

        qsa = """class ProveedorDefectoSetter {
        var esta_no;
        static var esta_si;
        static function fun1() {

            }
        function fun2() {
            }
        }"""

        valid = """# /** @class_declaration ProveedorDefectoSetter */
class ProveedorDefectoSetter(qsa.ObjectClass):
    # esta_no : Any = None
    esta_si: Any = None

    @classmethod
    def fun1(self):
        pass

    def fun2(self):
        pass
"""
        require_qs_py = qs2py(qsa)

        self.assertEqual(require_qs_py, valid)

    def test_require(self) -> None:
        """Test conveting fixture require.qs"""
        self.maxDiff = None
        require_qs = fixture_read("require.qs")

        old_tags = parser_qsa.IGNORE_NO_PYTHON_TAGS
        old_desktop = utils_base.FORCE_DESKTOP
        parser_qsa.IGNORE_NO_PYTHON_TAGS = False
        utils_base.FORCE_DESKTOP = False
        require_qs_py = qs2py(require_qs)
        require_py = fixture_read("require.python")

        # Delete version translator tag.
        pos_ini = require_qs_py.find("# Translated with pineboolib ")
        pos_fin = require_qs_py[pos_ini:].find("\n")
        require_qs_py = require_qs_py.replace(require_qs_py[pos_ini : pos_ini + pos_fin + 1], "")
        parser_qsa.IGNORE_NO_PYTHON_TAGS = old_tags
        utils_base.FORCE_DESKTOP = old_desktop
        self.assertEqual(require_qs_py, require_py)

    def test_staticclass(self) -> None:
        """Test static class."""

        self.maxDiff = None
        file_qs = fixture_read("static_class.qs")
        require_qs_py = qs2py(file_qs)
        require_py = fixture_read("static_class.python")

        # Delete version translator tag.
        pos_ini = require_qs_py.find("# Translated with pineboolib ")
        pos_fin = require_qs_py[pos_ini:].find("\n")
        require_qs_py = require_qs_py.replace(require_qs_py[pos_ini : pos_ini + pos_fin + 1], "")
        self.assertEqual(require_qs_py, require_py)

        """     def test_var_first(self) -> None:
        "Test when a var is first."

        self.maxDiff = None
        file_qs = fixture_read("varfirst.qs")
        require_qs_py = qs2py(file_qs)

        require_py = fixture_read("varfirst.python")

        # Delete version translator tag.
        pos_ini = require_qs_py.find("# Translated with pineboolib ")
        pos_fin = require_qs_py[pos_ini:].find("\n")
        require_qs_py = require_qs_py.replace(require_qs_py[pos_ini : pos_ini + pos_fin + 1], "")
        self.assertEqual(require_qs_py, require_py) """

    def test_dict(self) -> None:
        """Test dictionary."""

        qsa = """const _i = this.iface;

                const aDatosFactRect = {
                "codalmacen" : _i.calcularCodAlmacenFacturaRect(cursor, facturaRect),
                "regimeniva" : _i.calcularRegimenIvaFacturaRect(cursor, facturaRect),
                "codagente" : _i.calcularCodAgenteFacturaRect(cursor, facturaRect),
                }"""
        require_qs_py = qs2py(qsa)
        self.assertTrue(require_qs_py)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()



"""Aqobjectquerylist module."""

from PyQt6 import QtCore, QtWidgets  # type: ignore[import]

from pineboolib.application import types
from typing import Union, Optional


def aq_object_query_list(
    obj_: Union[QtCore.QObject, int, None],
    inherits_class: Optional[str] = None,
    object_name: Optional[str] = None,
    reg_ext_match: bool = False,
    recursirve_search: bool = False,
) -> types.Array:
    """Return a list with objects."""

    if isinstance(obj_, int) or obj_ is None:
        obj_ = QtWidgets.QApplication.topLevelWidgets()[0]

    if inherits_class is None:
        inherits_class = "QWidgets"

    class_ = getattr(QtWidgets, inherits_class, QtWidgets.QWidget)

    args_ = []

    args_.append(class_)

    if object_name and object_name == "":
        object_name = None

    args_.append(object_name)

    if recursirve_search:
        args_.append(QtCore.Qt.FindChildOption.FindChildrenRecursively)

    return types.Array(obj_.findChildren(*args_))  # type: ignore [arg-type]



"""Lineedit module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets  # type: ignore
from pineboolib.q3widgets import qwidget
from typing import Any


class LineEdit(qwidget.QWidget):
    """LineEdit class."""

    _label: QtWidgets.QLabel
    _line: QtWidgets.QLineEdit

    def __init__(self) -> None:
        """Inicialize."""

        super(LineEdit, self).__init__()

        self._label = QtWidgets.QLabel(self)
        self._line = QtWidgets.QLineEdit(self)
        _lay = QtWidgets.QHBoxLayout()
        _lay.addWidget(self._label)
        _lay.addWidget(self._line)
        self.setLayout(_lay)

    def __setattr__(self, name: str, value: Any) -> None:
        """Set attributes."""

        if name == "label":
            self._label.setText(str(value))
        elif name == "text":
            self._line.setText(str(value))
        else:
            super(LineEdit, self).__setattr__(name, value)

    def __getattr__(self, name: str) -> Any:
        """Return attribute."""

        if name == "text":
            return self._line.text()



# -*- coding: utf-8 -*-
"""dlgconnect module."""

import os
from PyQt6 import QtWidgets, QtGui, QtCore  # type: ignore[import]


from pineboolib import logging
from pineboolib.core import decorators, settings
from pineboolib.loader import projectconfig

from typing import Optional, cast, Dict, Any

LOGGER = logging.get_logger(__name__)


class DlgConnect(QtWidgets.QWidget):
    """
    DlgConnect Class.

    This class shows manages the Login dialog.
    """

    _options_showed: bool
    _min_size: "QtCore.QSize"
    _max_size: "QtCore.QSize"
    edit_mode: bool

    profiles: Dict[
        str, "projectconfig.ProjectConfig"
    ]  #: Index of loaded profiles. Keyed by description.
    selected_project_config: Optional[
        "projectconfig.ProjectConfig"
    ]  #: Contains the selected item to load.

    def __init__(self) -> None:
        """
        Initialize.
        """
        from pineboolib.application.database import pnsqldriversmanager

        super().__init__()
        self._options_showed = False
        self._min_size = QtCore.QSize(350, 140)
        self._max_size = QtCore.QSize(350, 495)
        self.profile_dir: str = projectconfig.ProjectConfig.profile_dir
        self.sql_drivers = pnsqldriversmanager.PNSqlDriversManager()
        self.edit_mode = False
        self.profiles = {}
        self.selected_project_config = None

    def load(self) -> None:
        """
        Load the dlgconnect form.
        """
        from pineboolib.fllegacy import flmanagermodules
        from pineboolib.core.utils import utils_base

        dlg_ = utils_base.filedir("loader/dlgconnect/dlgconnect.ui")

        self._user_interface: Any = flmanagermodules.FLManagerModules.createUI(dlg_, None, self)
        if not self._user_interface:
            raise Exception("Error creating dlgConnect")

        frame_geo = self.frameGeometry()
        primary_screen = QtGui.QGuiApplication.primaryScreen()
        if primary_screen:
            frame_geo.moveCenter(primary_screen.geometry().center())
        self.move(frame_geo.topLeft())

        self._user_interface.pbLogin.clicked.connect(self.open)
        self._user_interface.tbOptions.clicked.connect(self.toggleOptions)
        self._user_interface.pbSaveConnection.clicked.connect(self.saveProfile)
        self._user_interface.tbDeleteProfile.clicked.connect(self.deleteProfile)
        self._user_interface.tbEditProfile.clicked.connect(self.editProfile)
        self.cleanProfileForm()
        self._user_interface.cbDBType.currentIndexChanged.connect(self.updatePort)
        self._user_interface.cbProfiles.currentIndexChanged.connect(self.enablePassword)
        self._user_interface.cbAutoLogin.stateChanged.connect(self.cbAutoLogin_checked)
        self._user_interface.le_profiles.setText(self.profile_dir)
        self._user_interface.tb_profiles.clicked.connect(self.change_profile_dir)
        self.showOptions(False)
        self.loadProfiles()
        self._user_interface.leDescription.textChanged.connect(self.updateDBName)
        self._user_interface.installEventFilter(self)

    def cleanProfileForm(self) -> None:
        """
        Clean the profiles creation tab, and fill in the basic data of the default SQL driver.
        """
        self._user_interface.leDescription.setText("")
        driver_list = self.sql_drivers.aliasList()
        self._user_interface.cbDBType.clear()
        self._user_interface.cbDBType.addItems(driver_list)
        self._user_interface.cbDBType.setCurrentText(self.sql_drivers.defaultDriverName())
        self._user_interface.leURL.setText("localhost")
        self._user_interface.leDBUser.setText("")
        self._user_interface.leDBPassword.setText("")
        self._user_interface.leDBName.setText("")
        self._user_interface.leProfilePassword.setText("")
        self._user_interface.leProfilePassword2.setText("")
        self._user_interface.cbAutoLogin.setChecked(False)
        self.updatePort()

    def loadProfiles(self) -> None:
        """
        Update ComboBox of profiles.
        """
        if not os.path.exists(self.profile_dir):
            return

        self._user_interface.cbProfiles.clear()
        self.profiles.clear()

        with os.scandir(self.profile_dir) as profile:
            for entry in profile:
                if entry.name.startswith("."):
                    continue
                if not entry.name.endswith(".xml"):
                    continue
                if not entry.is_file():
                    continue

                pconf = projectconfig.ProjectConfig(
                    filename=os.path.join(self.profile_dir, entry.name),
                    database="unset",
                    type="unset",
                )
                try:
                    pconf.load_projectxml()
                except projectconfig.PasswordMismatchError:
                    LOGGER.trace(
                        "Profile %r [%r] requires a password", pconf.description, entry.name
                    )
                except Exception:
                    LOGGER.exception("Unexpected error trying to read profile %r", entry.name)
                    continue
                self.profiles[pconf.description] = pconf

        for name in sorted(self.profiles.keys()):
            self._user_interface.cbProfiles.addItem(name)

        last_profile = settings.SETTINGS.value("DBA/last_profile", None)

        if last_profile:
            self._user_interface.cbProfiles.setCurrentText(last_profile)

    @decorators.pyqt_slot()
    def toggleOptions(self) -> None:
        """Show/Hide Options."""
        self.showOptions(not self._options_showed)

    def showOptions(self, show_options: bool) -> None:
        """
        Show the frame options.
        """
        if show_options:
            self._user_interface.frmOptions.show()
            self._user_interface.tbDeleteProfile.show()
            self._user_interface.tbEditProfile.show()
            self._user_interface.setMinimumSize(self._max_size)
            self._user_interface.setMaximumSize(self._max_size)
            self._user_interface.resize(self._max_size)
        else:
            self._user_interface.frmOptions.hide()
            self._user_interface.tbDeleteProfile.hide()
            self._user_interface.tbEditProfile.hide()
            self._user_interface.setMinimumSize(self._min_size)
            self._user_interface.setMaximumSize(self._min_size)
            self._user_interface.resize(self._min_size)

        self._options_showed = show_options

    @decorators.pyqt_slot()
    def open(self) -> None:
        """
        Open the selected connection.
        """
        current_profile = self._user_interface.cbProfiles.currentText()
        pconf = self.getProjectConfig(current_profile)
        if pconf is None:
            return
        self.selected_project_config = pconf
        settings.SETTINGS.set_value("DBA/last_profile", current_profile)
        self.close()

    @decorators.pyqt_slot()
    def saveProfile(self) -> None:
        """
        Save the connection.
        """
        if self._user_interface.leDescription.text() == "":
            QtWidgets.QMessageBox.information(
                self._user_interface, "Pineboo", "La descripción no se puede dejar en blanco"
            )
            self._user_interface.leDescription.setFocus()
            return

        if self._user_interface.leDBPassword.text() != self._user_interface.leDBPassword2.text():
            QtWidgets.QMessageBox.information(
                self._user_interface, "Pineboo", "La contraseña de la BD no coincide"
            )
            self._user_interface.leDBPassword.setText("")
            self._user_interface.leDBPassword2.setText("")
            return

        if (
            self._user_interface.leProfilePassword.text()
            != self._user_interface.leProfilePassword2.text()
        ):
            QtWidgets.QMessageBox.information(
                self._user_interface, "Pineboo", "La contraseña del perfil no coincide"
            )
            self._user_interface.leProfilePassword.setText("")
            self._user_interface.leProfilePassword2.setText("")
            return

        if self.edit_mode:
            pconf = self.getProjectConfig(self._user_interface.cbProfiles.currentText())
            if pconf is None:
                return
            pconf.description = self._user_interface.leDescription.text()
        else:
            pconf = projectconfig.ProjectConfig(
                description=self._user_interface.leDescription.text(),
                database="unset",
                type="unset",
            )

        if not os.path.exists(self.profile_dir):
            import pathlib

            pathlib.Path(self.profile_dir).mkdir(parents=True, exist_ok=True)

        if os.path.exists(pconf.filename) and not self.edit_mode:
            QtWidgets.QMessageBox.information(
                self._user_interface, "Pineboo", "El perfil ya existe"
            )
            return

        pconf.type = self._user_interface.cbDBType.currentText()
        pconf.host = self._user_interface.leURL.text()
        pconf.port = int(self._user_interface.lePort.text())
        pconf.username = self._user_interface.leDBUser.text()
        pconf.password = self._user_interface.leDBPassword.text()
        pconf.database = self._user_interface.leDBName.text()

        pass_profile_text = ""
        if not self._user_interface.cbAutoLogin.isChecked():
            pass_profile_text = self._user_interface.leProfilePassword.text()
        pconf.project_password = pass_profile_text
        pconf.save_projectxml(overwrite_existing=self.edit_mode)
        # self.cleanProfileForm()
        self.loadProfiles()
        self._user_interface.cbProfiles.setCurrentText(pconf.description)

    @decorators.pyqt_slot()
    def deleteProfile(self) -> None:
        """
        Delete the selected connection.
        """
        if self._user_interface.cbProfiles.count() > 0:
            res = QtWidgets.QMessageBox.warning(
                self._user_interface,
                "Pineboo",
                "¿Desea borrar el perfil %s?" % self._user_interface.cbProfiles.currentText(),
                cast(
                    QtWidgets.QMessageBox.StandardButton,
                    QtWidgets.QMessageBox.StandardButton.Ok
                    | QtWidgets.QMessageBox.StandardButton.No,
                ),
                QtWidgets.QMessageBox.StandardButton.No,
            )
            if res == QtWidgets.QMessageBox.StandardButton.No:
                return

            pconf: "projectconfig.ProjectConfig" = self.profiles[
                self._user_interface.cbProfiles.currentText()
            ]
            os.remove(pconf.filename)
            self.loadProfiles()

    def getProjectConfig(self, name: str) -> Optional["projectconfig.ProjectConfig"]:
        """
        Get a profile by name and ensure its fully loaded.
        """
        if name not in self.profiles.keys():
            return None

        pconf: "projectconfig.ProjectConfig" = self.profiles[name]

        if pconf.password_required:
            # As it failed to load earlier, it needs a password.
            # Copy the current password and test again...
            pconf.project_password = self._user_interface.lePassword.text()
            try:
                pconf.load_projectxml()
            except projectconfig.PasswordMismatchError:
                QtWidgets.QMessageBox.information(
                    self._user_interface, "Pineboo", "Contraseña Incorrecta"
                )
                return None
        return pconf

    @decorators.pyqt_slot()
    def editProfile(self) -> None:
        """
        Edit the selected connection.
        """
        # Cogemos el perfil y lo abrimos
        self.editProfileName(self._user_interface.cbProfiles.currentText())

    def editProfileName(self, name: str) -> None:
        """
        Edit profile from name. Must have been loaded earlier on loadProfiles.
        """
        pconf = self.getProjectConfig(name)
        if pconf is None:
            return

        self._user_interface.leProfilePassword.setText(pconf.project_password)
        self._user_interface.leProfilePassword2.setText(pconf.project_password)

        self._user_interface.cbAutoLogin.setChecked(pconf.project_password == "")

        self._user_interface.leDescription.setText(pconf.description)
        self._user_interface.leDBName.setText(pconf.database)

        self._user_interface.leURL.setText(pconf.host)
        self._user_interface.lePort.setText(str(pconf.port))
        self._user_interface.cbDBType.setCurrentText(pconf.type)

        self._user_interface.leDBUser.setText(pconf.username)

        self._user_interface.leDBPassword.setText(pconf.password)
        self._user_interface.leDBPassword2.setText(pconf.password)

        self.edit_mode = True

    @decorators.pyqt_slot(int)
    def updatePort(self) -> None:
        """
        Update to the driver default port.
        """
        self._user_interface.lePort.setText(
            self.sql_drivers.port(self._user_interface.cbDBType.currentText())
        )

    @decorators.pyqt_slot(int)
    def enablePassword(self, enable: Optional[int] = None) -> None:
        """
        Check if the profile requires password to login or not.
        """
        if self._user_interface.cbProfiles.count() == 0:
            return
        pconf: "projectconfig.ProjectConfig" = self.profiles[
            self._user_interface.cbProfiles.currentText()
        ]
        # NOTE: This disables the password entry once the password has been processed for
        # .. the profile once. So the user does not need to retype it.
        self._user_interface.lePassword.setEnabled(pconf.password_required)
        self._user_interface.lePassword.setText("")

    def updateDBName(self) -> None:
        """
        Update the name of the database with the description name.
        """
        self._user_interface.leDBName.setText(
            self._user_interface.leDescription.text().replace(" ", "_").lower()
        )

    @decorators.pyqt_slot(int)
    def cbAutoLogin_checked(self) -> None:
        """
        Process checked event from AutoLogin checkbox.
        """

        if self._user_interface.cbAutoLogin.isChecked():
            self._user_interface.leProfilePassword.setEnabled(False)
            self._user_interface.leProfilePassword2.setEnabled(False)
        else:
            self._user_interface.leProfilePassword.setEnabled(True)
            self._user_interface.leProfilePassword2.setEnabled(True)

    def change_profile_dir(self) -> None:
        """
        Change the path where profiles are saved.
        """

        new_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self._user_interface,
            self.tr("Carpeta profiles"),
            self.profile_dir,
            QtWidgets.QFileDialog.Option.ShowDirsOnly,
        )

        if new_dir and new_dir is not self.profile_dir:
            settings.CONFIG.set_value("ebcomportamiento/profiles_folder", new_dir)
            self.profile_dir = new_dir
            projectconfig.ProjectConfig.profile_dir = new_dir
            self.loadProfiles()

    def eventFilter(
        self, object: Optional["QtCore.QObject"], event: Optional["QtCore.QEvent"]
    ) -> bool:
        """Event Filter."""

        if isinstance(event, QtGui.QKeyEvent):
            if event.key() in (
                cast(int, QtCore.Qt.Key.Key_Return),
                cast(int, QtCore.Qt.Key.Key_Enter),
            ):
                self.open()
                return True

            elif event.key() == cast(int, QtCore.Qt.Key.Key_Escape):
                self.close()
                return True

        return super().eventFilter(object, event)  # type: ignore [arg-type]



"""Fltable module."""

# -*- coding: utf-8 -*-
from pineboolib.q3widgets import qtable
from pineboolib.core import decorators
from typing import Any


class FLTable(qtable.QTable):
    """FLTable class."""

    AlwaysOff: bool

    @decorators.not_implemented_warn
    def setColumnMovingEnabled(self, value: bool) -> None:
        """Set that columns can be moved."""
        pass

    @decorators.not_implemented_warn
    def setVScrollBarMode(self, mode: Any) -> None:
        """Set the vertical scroll bar mode."""
        pass

    @decorators.not_implemented_warn
    def setHScrollBarMode(self, mode: Any) -> None:
        """Set the horizontal scroll bar mode."""
        pass



"""Test_sysbasetype module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.application.qsatypes import sysbasetype
from pineboolib import application

from typing import Any


class TestSysBaseClassGeneral(unittest.TestCase):
    """TestSysBaseClassGeneral Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test basic functions."""
        import platform
        from pineboolib.core.utils.utils_base import filedir

        import os

        base_type = sysbasetype.SysBaseType()
        self.assertEqual(base_type.nameUser(), "memory_user")
        # self.assertEqual(base_type.interactiveGUI(), "Pineboo")
        self.assertEqual(base_type.isLoadedModule("sys"), True)
        os_name = "LINUX"
        if platform.system() == "Windows":
            os_name = "WIN32"
        elif platform.system() == "Darwin":
            os_name = "MACX"

        self.assertEqual(base_type.osName(), os_name)
        self.assertEqual(base_type.nameBD(), "temp_db")
        self.assertEqual(base_type.installPrefix(), filedir(".."))
        self.assertEqual(base_type.version(), str(application.PROJECT.load_version()))
        file_path = "%s/test_sysbasetype.txt" % application.PROJECT.tmpdir

        if os.path.exists(file_path):
            os.remove(file_path)

        base_type.write("ISO-8859-15", file_path, "avión, caña")
        self.assertEqual(os.path.exists(file_path), True)
        self.assertEqual(base_type.nameDriver(), "FLsqlite")
        self.assertEqual(base_type.nameHost(), "")

    def test_basic_2(self) -> None:
        """Test Basic functions."""
        base_type = sysbasetype.SysBaseType()
        self.assertFalse(base_type.isDeveloperBuild())
        self.assertFalse(base_type.isNebulaBuild())
        self.assertFalse(base_type.isCloudMode())

        base_type.Mr_Proper()
        base_type.cleanupMetaData()

    def test_basic_3(self) -> None:
        """Test basic functions."""
        base_type = sysbasetype.SysBaseType()
        report_changes = "Nombre: 1 \nEstado: 2 \nShaOldTxt: 3 \nShaNewTxt: 4 \n###########################################\n"
        self.assertEqual(base_type.reportChanges({"prueba": "1@2@3@4"}), report_changes)

    def test_objects(self) -> None:
        """Test objects functions."""
        from pineboolib.fllegacy import fltabledb

        application.PROJECT.actions["flareas"].openDefaultForm()

        form = application.PROJECT.actions[  # type: ignore [attr-defined] # noqa F821
            "flareas"
        ]._master_widget
        # form = flformdb.FLFormDB(None, action)
        # self.assertTrue(form)
        # form.load()
        if form is None:
            self.assertTrue(form)  # type: ignore [unreachable] # noqa F821
            return

        fltable: Any = form.findChild(fltabledb.FLTableDB, "tableDBRecords")

        base_type = sysbasetype.SysBaseType()

        self.assertFalse(base_type.setObjText(fltable, "flfielddb_5", "Holas"))
        self.assertFalse(base_type.setObjText(fltable, "toolButtonInsert", "prueba"))

        base_type.disableObj(form, "tableDBRecords")
        base_type.enableObj(form, "tableDBRecords")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestSysBaseClassDataBase(unittest.TestCase):
    """TestSysBaseClassDataBase Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test addDatabase and removeDatabase functions."""

        base_type = sysbasetype.SysBaseType()

        prueba_conn_1 = application.PROJECT.conn_manager.useConn("prueba")
        self.assertEqual(prueba_conn_1.isOpen(), False)
        self.assertEqual(base_type.addDatabase("prueba"), True)
        self.assertEqual(prueba_conn_1.isOpen(), True)
        self.assertEqual(base_type.removeDatabase("prueba"), True)
        self.assertNotEqual(base_type.idSession(), None)
        self.assertEqual(prueba_conn_1.isOpen(), False)
        prueba_conn_2 = application.PROJECT.conn_manager.useConn("prueba")
        self.assertEqual(prueba_conn_2.isOpen(), False)
        self.assertEqual(base_type.addDatabase("prueba"), True)
        self.assertEqual(prueba_conn_2.isOpen(), True)
        self.assertEqual(base_type.removeDatabase("prueba"), True)
        self.assertEqual(prueba_conn_1.isOpen(), False)

        self.assertTrue(
            base_type.addDatabase(
                "FLsqlite",
                ":memory:",
                prueba_conn_1.user(),
                prueba_conn_1.returnword(),
                prueba_conn_1.host(),
                prueba_conn_1.port(),
                "extra",
            )
        )
        self.assertEqual(application.PROJECT.conn_manager.useConn("extra")._db_name, ":memory:")
        self.assertEqual(
            application.PROJECT.conn_manager.useConn("extra")._db_host, prueba_conn_1.host()
        )
        self.assertEqual(
            application.PROJECT.conn_manager.useConn("extra")._db_port, prueba_conn_1.port()
        )

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


if __name__ == "__main__":
    unittest.main()



"""Test loadmodpkg module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa


class TestFLLoadModPkg(unittest.TestCase):
    """TestFLLoadModPkg class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""

        init_testing()

    def test_basic1(self) -> None:
        """Test loadmodpkg main function."""

        _func = qsa.from_project("formflloadmodpkg").main
        self.assertTrue(_func)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()



"""Test projectmodule module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.application.tests import fixture_path


class TestProjectModule(unittest.TestCase):
    """TestProjectModule Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_parse_script(self) -> None:
        """Test parse_script function."""

        from pineboolib import application
        import os
        import shutil

        path = fixture_path("flfacturac.qs")
        tmp_path = "%s/%s" % (application.PROJECT.tmpdir, "temp_qs_projectmodule.qs")
        path_py = "%s.qs.py" % tmp_path[:-3]
        if os.path.exists(tmp_path):
            os.remove(tmp_path)

        if os.path.exists(path_py):
            os.remove(path_py)

        shutil.copy(path, tmp_path)
        application.PROJECT.parse_script(tmp_path)

        self.assertTrue(os.path.exists(path_py))

        if os.path.exists(tmp_path):
            os.remove(tmp_path)

        if os.path.exists(path_py):
            os.remove(path_py)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Test consistency module."""

import unittest
import threading

from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import application
from pineboolib.core.utils import utils_base
from pineboolib.qsa import qsa


class TestConsistency(unittest.TestCase):
    """TestConsistency Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_isolation(self) -> None:
        """Create multiples diferents sessions."""

        conn_name = "default"

        single_session = qsa.session(conn_name)
        legacy_session = qsa.session(conn_name, True)
        thread_session = qsa.thread_session_new(conn_name)

        self.assertTrue(single_session is not legacy_session)
        self.assertTrue(thread_session is not single_session)
        self.assertEqual(
            legacy_session, application.PROJECT.conn_manager.useConn("default").session()
        )

    @qsa.serialize()  # type: ignore [misc] # noqa: F821
    def test_serialize(self) -> None:
        """Test serialize decorator."""
        conn_ident = utils_base.session_id()
        id_thread = threading.current_thread().ident
        self.assertTrue(
            conn_ident in application.SERIALIZE_LIST[id_thread]  # type: ignore [index] # noqa: F821
        )

    @qsa.atomic()  # type: ignore [misc] # noqa: F821
    def test_transaction(self) -> None:
        """Create a new record and query it from a query in the same transaction."""
        self.assertTrue(atomica())
        session = qsa.session_atomic()
        self.assertTrue(session)

        class_ = qsa.orm.fltest
        obj_1 = class_()
        self.assertFalse(obj_1.string_field)
        self.assertFalse(obj_1.empty_relation)
        obj_1.empty_relation = None
        self.assertFalse(obj_1.empty_relation)
        self.assertTrue(obj_1.save())
        self.assertTrue(obj_1.id)

        cursor_fltest = qsa.FLSqlCursor("fltest")
        cursor_fltest.select("id = %s" % obj_1.id)
        self.assertTrue(cursor_fltest.first())

        self.assertTrue(session is obj_1.session)
        self.assertTrue(session is cursor_fltest.db().session())

        # Check string_field
        result = qsa.FLUtil.sqlSelect("fltest", "string_field", "id = %s" % obj_1.id)
        self.assertFalse(result)
        self.assertTrue(result == "", 'El valor devuelto (%s) no es ""' % result)

        self.assertTrue(
            cursor_fltest.valueBuffer("string_field") == "",
            'El valor devuelto (%s) no es ""' % result,
        )
        # Check empty_relation
        self.assertTrue(obj_1.empty_relation is None)
        result_er = qsa.FLUtil.sqlSelect("fltest", "empty_relation", "id = %s" % obj_1.id)
        self.assertFalse(result_er)
        self.assertTrue(result_er == "", 'El valor devuelto (%s) no es ""' % result_er)

        self.assertTrue(
            cursor_fltest.valueBuffer("empty_relation") == "",
            'El valor devuelto (%s) no es ""' % result,
        )

    def test_save_point_launch(self) -> None:
        """Test save points."""

        conn_ = qsa.aqApp.db().useConn("default")
        session = conn_.session()
        self.assertTrue(session.transaction is None)
        conn_.transaction()
        self.assertTrue(session.transaction is not None)
        id_ = session.transaction
        conn_.transaction()
        self.assertNotEqual(id_, session.transaction)
        self.assertTrue(id_ is not None)
        conn_.rollback()
        self.assertEqual(id_, session.transaction)
        conn_.rollback()
        self.assertTrue(session.transaction is None)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()


def atomica():
    """Atomica function test."""
    obj_area = qsa.orm.flareas()
    obj_area.idarea = "A"
    obj_area.descripcion = "Area A"
    obj_area.save()
    qry = qsa.FLUtil.sqlSelect("flareas", "descripcion", "idarea = 'A'")
    return qry == "Area A" and qsa.session_atomic() is not None



"""Test_pnaction module."""
import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.application.metadata import pnaction
from pineboolib import application
from pineboolib.qsa import qsa


class TestPNAction(unittest.TestCase):
    """TestPNAction class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic1(self) -> None:
        """Test basic 1."""

        qsa.sys.reinit()
        act_struct = application.PROJECT.actions["flareas"]
        self.assertTrue(act_struct is not None)
        act_ = pnaction.PNAction(act_struct)
        self.assertEqual(act_.scriptFormRecord(), "")
        self.assertEqual(act_.scriptForm(), "flmasterareas.qs")
        self.assertEqual(act_.table(), "flareas")
        self.assertEqual(act_.formRecord(), "flareas.ui")
        self.assertEqual(act_.form(), "master.ui")
        act2_ = act_
        self.assertTrue(act2_ == act_)
        self.assertTrue(str(act_) is not None)
        self.assertFalse(act_.class_())

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Flmrproper module."""
# -*- coding: utf-8 -*-
from pineboolib.qsa import qsa


class FormInternalObj(qsa.FormDBWidget):
    """FormInternalObj class."""

    def main(self) -> None:
        """Entry function."""
        qsa.sys.Mr_Proper()



"""
To resolve file and folder paths.
"""

from pineboolib.core.utils import logging
from pineboolib import application

import os
from typing import Optional, List

LOGGER = logging.get_logger(__name__)


def _dir(*x) -> str:
    """
    Calculate the path of a folder.

    @param x. str or array with the folder path.
    @return str with absolute path to a folder.
    """

    list_: List[str] = list(x)
    if os.name == "nt":
        list_ = [item.replace("/", "\\") for item in list_]

    return os.path.join(application.PROJECT.tmpdir, *list_)


def coalesce_path(*filenames) -> Optional[str]:
    """
    Return the first existing file in a group of files.

    @return path to the first file found.
    """
    for filename in filenames:
        if filename is None:
            # When the caller specifies None as the last item means that its OK to return None
            return None

        if filename in application.PROJECT.files:
            return application.PROJECT.files[filename].path()
    LOGGER.error(
        "coalesce_path: Ninguno de los ficheros especificados ha sido encontrado en el proyecto: %s",
        repr(filenames),
        stack_info=False,
    )
    return None


def _path(filename: str, show_not_found: bool = True) -> Optional[str]:
    """
    Return the first existing file in a group of files.

    @return path to file.
    """
    if filename not in application.PROJECT.files:
        if show_not_found:
            LOGGER.error("Fichero %s no encontrado en el proyecto.", filename, stack_info=False)
        return None
    return application.PROJECT.files[filename].path()



"""Test_pnaccesscontrol module."""

from PyQt6 import QtXml  # type: ignore[import]

import unittest
from pineboolib.loader.main import init_testing, finish_testing


from pineboolib.application.acls import pnaccesscontrol


class TestPNAccessControl(unittest.TestCase):
    """TestPNAccessControl Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_missing_lines(self) -> None:
        """Test missing lines."""

        ac_ = pnaccesscontrol.PNAccessControl()
        ac_._acos_perms["uno"] = "uno:dos:tres"
        self.assertFalse(ac_.type())
        ac_.set(None)  # type: ignore [arg-type] # noqa: F821
        ac_.set(QtXml.QDomElement())
        ac_.get(None)  # type: ignore [arg-type] # noqa: F821
        ac_.get(QtXml.QDomDocument())

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Dialog module."""

from PyQt6 import QtCore, QtWidgets  # type: ignore[import]

from pineboolib.q3widgets import qdialog
from pineboolib.q3widgets import qpushbutton
from pineboolib.q3widgets import qtabwidget

from typing import Optional, Union


class Dialog(qdialog.QDialog):
    """Dialog class."""

    _layout: QtWidgets.QVBoxLayout
    _button_box: QtWidgets.QDialogButtonBox
    okButton: qpushbutton.QPushButton
    cancelButton: qpushbutton.QPushButton
    _tab: qtabwidget.QTabWidget

    def __init__(
        self,
        title: Optional[str] = None,
        f: Union[Optional[QtCore.Qt.WindowType], int] = None,
        desc: Optional[str] = None,
    ) -> None:
        """Inicialize."""

        # FIXME: f no lo uso , es qt.windowsflg
        super(Dialog, self).__init__(None, f if isinstance(f, str) else None)
        if title:
            self.setWindowTitle(str(title))

        self.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
        self._layout = QtWidgets.QVBoxLayout()
        self.setLayout(self._layout)
        self._button_box = QtWidgets.QDialogButtonBox()
        self.okButton = qpushbutton.QPushButton("&Aceptar")  # pylint: disable=invalid-name
        self.cancelButton = qpushbutton.QPushButton("&Cancelar")  # pylint: disable=invalid-name
        self.okButtonText = "Aceptar"  # pylint: disable=invalid-name
        self.cancelButtonText = "Cancelar"  # pylint: disable=invalid-name

        self._button_box.addButton(self.okButton, QtWidgets.QDialogButtonBox.ButtonRole.AcceptRole)
        self._button_box.addButton(
            self.cancelButton, QtWidgets.QDialogButtonBox.ButtonRole.RejectRole
        )
        self.okButton.clicked.connect(self.accept)  # type: ignore [attr-defined] # noqa: F821
        self.cancelButton.clicked.connect(self.reject)  # type: ignore [attr-defined] # noqa: F821
        self._tab = qtabwidget.QTabWidget()
        self._tab.hide()

    def add(self, _object: QtWidgets.QWidget) -> None:
        """Add widget to Dialog."""

        self._layout.addWidget(_object)

    def newTab(self, name: str) -> None:
        """Add a new tab to Dialog."""

        if self._tab.isHidden():
            self._tab.show()
        self._tab.addTab(QtWidgets.QWidget(), str(name))

    def setWidth(self, width: Union[int, float]) -> None:
        """Set width."""

        height = self.height()
        # self.setMinimunSize(width, height)
        self.resize(int(width), height)

    def exec_(self) -> int:
        """Execute Dialog."""

        self._layout.addWidget(self._button_box)
        return super().exec()

    def setHeight(self, height: Union[int, float]) -> None:
        """Set height."""

        width = self.width()
        # self.setMinimunSize(width, height)
        self.resize(width, int(height))

    def getCancelButtonText(self) -> str:
        """Return cancel button text."""

        return self.cancelButton.text

    def setCancelButtonText(self, text_: str) -> None:
        """Set cancel button text."""

        self.cancelButton.setText(text_)

    def getOkButtonText(self) -> str:
        """Return cancel button text."""

        return self.okButton.text

    def setOkButtonText(self, text_: str) -> None:
        """Set cancel button text."""

        self.okButton.setText(text_)

    def getTitle(self) -> str:
        """Return dialog title."""

        return self.windowTitle()

    def setTitle(self, title_: str) -> None:
        """Set dialog title."""

        self.setWindowTitle(title_)

    cancelButtonText: str = property(  # type: ignore [assignment] # noqa: F821
        getCancelButtonText, setCancelButtonText
    )
    okButtonText: str = property(  # type: ignore [assignment] # noqa: F821
        getOkButtonText, setOkButtonText
    )
    title: str = property(getTitle, setTitle)  # type: ignore [assignment] # noqa: F821



"""Qbuttongroup module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets, QtCore  # type: ignore[import]
from pineboolib.q3widgets import qgroupbox


from typing import Callable, Optional


class QButtonGroup(qgroupbox.QGroupBox):
    """QButtonGroup class."""

    pressed = QtCore.pyqtSignal(int)  # type: ignore [assignment] # noqa: F821
    clicked = QtCore.pyqtSignal(int)  # type: ignore [assignment] # noqa: F821

    def __init__(self, *args) -> None:
        """Inicialize."""

        super().__init__(*args)
        self.bg_ = QtWidgets.QButtonGroup(self)
        self.selectedId = -1  # pylint: disable=invalid-name

    def setSelectedId(self, id_: int) -> None:
        """Set selected id."""

        self.selectedId = id_

    def __getattr__(self, name: str) -> Optional[Callable]:
        """Return an attribute."""

        ret_ = getattr(self.bg_, name, None)
        return ret_



"""PNCore class is used for receive signal from mainforms."""
from PyQt6 import QtCore, QtWidgets


class PNCore(QtWidgets.QWidget):
    """PNCore class."""

    @QtCore.pyqtSlot()
    def execDefaultScript(self):
        """Execdefaultscript class."""
        pass

    @QtCore.pyqtSlot()
    def openDefaultForm(self):
        """Opendefaultform class."""
        pass



"""Flfiles module."""
# -*- coding: utf-8 -*-
from pineboolib.qsa import qsa


class FormInternalObj(qsa.FormDBWidget):
    """FormInternalObj class."""

    def init(self) -> None:
        """Init function."""
        fdb_contenido = self.child("contenido")
        fdb_nombre = self.child("nombre")
        fdb_sha = self.child("sha")

        if fdb_contenido is None:
            raise Exception("contenido control not found!.")

        fdb_contenido.setText(self.cursor().valueBuffer("contenido"))
        fdb_nombre.setValue(self.cursor().valueBuffer("nombre"))
        fdb_sha.setValue(self.cursor().valueBuffer("sha"))

        edit_button = self.child("botonEditar")
        edit_pb_xml = self.child("pbXMLEditor")

        if edit_button:
            edit_button.setEnabled(False)

        if edit_pb_xml:
            edit_pb_xml.setEnabled(False)



"""Flsmtpclient module."""
# -*- coding: utf-8 -*-
from PyQt6 import QtCore, QtGui  # type: ignore
from os.path import basename
from pineboolib import logging

import smtplib
import socket

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.image import MIMEImage

from typing import List, Optional, Any, Dict

LOGGER = logging.get_logger(__name__)


class State(object):
    """State class."""

    Init: int = 0
    Mail: int = 1
    Rcpt: int = 2
    Data: int = 3
    Body: int = 4
    Quit: int = 5
    Close: int = 6
    SmtpError: int = 7
    Connecting: int = 8
    Connected: int = 9
    MxDnsError: int = 10
    SendOk: int = 11
    SockedError: int = 12
    Composing: int = 13
    Attach: int = 14
    AttachError: int = 15
    ServerError: int = 16
    ClientError: int = 17
    StartTTLS: int = 18
    WaitingForSTARTTLS: int = 19
    SendAuthPlain: int = 20
    SendAuthLogin: int = 21
    WaitingForAuthPlain: int = 22
    WaitingForAuthLogin: int = 23
    WaitingForUser: int = 24
    WaitingForPass: int = 25


class AuthMethod(object):
    """AuthMethod class."""

    NoAuth: int = 0
    AuthPlain: int = 1
    AuthLogin: int = 2


class ConnectionType(object):
    """ConnectionType class."""

    TcpConnection: int = 0
    SslConnection: int = 1
    TlsConnection: int = 2


class FLSmtpClient(QtCore.QObject, AuthMethod, ConnectionType, State):
    """FLSmtpClient class."""

    from_value_: Optional[str]
    reply_to_: Optional[str]
    to_: Optional[str]
    cc_: Optional[str]
    bcc_: Optional[str]
    organization_: Optional[str]
    priority_: int
    subject_: Optional[str]
    body_: Optional[str]

    attachments_: List[str]

    mail_server_: Optional[str]
    mime_type_: Optional[str]
    port_: Optional[int]

    text_parts_: List[str]
    map_attach_cid_: Dict[str, Any]

    status_msg_: Optional[str]
    state_code_: int

    user_: Optional[str]
    password_: Optional[str]
    connection_type_: int
    auth_method_: int

    status = QtCore.pyqtSignal(str)
    sendStarted = QtCore.pyqtSignal()
    sendEnded = QtCore.pyqtSignal()
    sendTotalSteps = QtCore.pyqtSignal(int)
    sendStepNumber = QtCore.pyqtSignal(int)
    statusChanged = QtCore.pyqtSignal(str, int)

    def __init__(self, parent: Optional[QtCore.QObject] = None) -> None:
        """Inicialize."""
        super(FLSmtpClient, self).__init__(parent)
        self.state_code_ = State.Init
        self.status_msg_ = None
        self.priority_ = 0
        self.port_ = 25
        self.connection_type_ = ConnectionType.TcpConnection
        self.auth_method_ = AuthMethod.NoAuth
        self.from_value_ = None
        self.reply_to_ = None
        self.to_ = None
        self.cc_ = None
        self.bcc_ = None
        self.organization_ = None
        self.priority_ = 0
        self.subject_ = None
        self.body_ = None
        self.attachments_ = []
        self.mail_server_ = None
        self.mime_type_ = None
        self.port_ = None
        self.text_parts_ = []
        self.map_attach_cid_ = {}
        self.user_ = None
        self.password_ = None
        self.connection_type_ = 0
        self.auth_method_ = 0

    def setFrom(self, from_: str) -> None:
        """Set from."""
        self.from_value_ = from_

    def from_(self) -> Optional[str]:
        """Return from."""

        return self.from_value_

    def setReplyTo(self, reply_to: str) -> None:
        """Set reply to."""

        self.reply_to_ = reply_to

    def replyTo(self) -> Optional[str]:
        """Return reply to."""

        return self.reply_to_

    def setTo(self, to_: str) -> None:
        """Set to."""
        self.to_ = to_

    def to(self) -> Optional[str]:
        """Return to."""
        return self.to_

    def setCC(self, cc_: str) -> None:
        """Set cc."""
        self.cc_ = cc_

    def CC(self) -> Optional[str]:
        """Return cc."""
        return self.cc_

    def setBCC(self, cc_: str) -> None:
        """Set bcc."""

        self.bcc_ = cc_

    def BCC(self) -> Optional[str]:
        """Return bcc."""

        return self.bcc_

    def setOrganization(self, org: str) -> None:
        """Set organization."""

        self.organization_ = org

    def organization(self) -> Optional[str]:
        """Return organization."""

        return self.organization_

    def setPriority(self, prio: int) -> None:
        """Set priority."""

        self.priority_ = prio

    def priority(self) -> int:
        """Return priority."""
        return self.priority_

    def setSubject(self, subject: str) -> None:
        """Set subject."""
        self.subject_ = subject

    def subject(self) -> Optional[str]:
        """Return subject."""

        return self.subject_

    def setBody(self, body: str) -> None:
        """Set body."""
        self.body_ = body

    def body(self) -> Optional[str]:
        """Return body."""
        return self.body_

    def addAttachment(self, attach: str, cid: Optional[Any] = None) -> None:
        """Add attachment file to mail."""
        if QtCore.QFile.exists(attach) and QtCore.QFileInfo(attach).isReadable():
            if attach and attach not in self.attachments_:
                self.attachments_.append(attach)
                if cid:
                    self.map_attach_cid_[attach] = cid
        else:
            err_msg_ = self.tr("El fichero %s no existe o no se puede leer\n\n" % attach)
            LOGGER.warning(err_msg_)
            self.changeStatus(err_msg_, State.AttachError)

    def addTextPart(self, text: Optional[str], mime_type: str = "text/plain") -> None:
        """Add text part to mail."""
        if text is not None:
            self.text_parts_.append(text)
            self.text_parts_.append(mime_type)

    def setMailServer(self, mail_server: str) -> None:
        """Set mail server."""
        self.mail_server_ = mail_server

    def mailServer(self) -> Optional[str]:
        """Return mail server."""
        return self.mail_server_

    def setMimeType(self, mine_type: str) -> None:
        """Set mine type."""
        self.mime_type_ = mine_type

    def mimeType(self) -> Optional[str]:
        """Return mine type."""

        return self.mime_type_

    def setPort(self, port: int) -> None:
        """Set port."""
        self.port_ = port

    def port(self) -> Optional[int]:
        """Return port."""
        return self.port_

    def lastStatusMsg(self) -> Optional[str]:
        """Return last status message."""

        return self.status_msg_

    def lastStateCode(self) -> int:
        """Return last state code."""
        return self.state_code_

    def setUser(self, user: str) -> None:
        """Set user name."""
        self.user_ = user

    def user(self) -> Optional[str]:
        """Return user name."""
        return self.user_

    def setPassword(self, password: str) -> None:
        """Set password."""
        self.password_ = password

    def password(self) -> Optional[str]:
        """Return password."""
        return self.password_

    def setConnectionType(self, conn: int) -> None:
        """Set connection type."""
        self.connection_type_ = conn

    def connectionType(self) -> int:
        """Return connection type."""

        return self.connection_type_

    def setAuthMethod(self, method: int) -> None:
        """Set authentication method."""
        self.auth_method_ = method

    def authMethod(self) -> int:
        """Return authentication method."""
        return self.auth_method_

    def startSend(self) -> bool:
        """Start send mail."""

        from pineboolib.core.utils import utils_base
        from pineboolib.core import settings
        from pineboolib import application

        self.sendStarted.emit()
        self.sendTotalSteps.emit(len(self.attachments_) + 3)

        step = 0

        self.changeStatus(self.tr("Componiendo mensaje"), State.Composing)

        outer = MIMEMultipart()
        outer["From"] = self.from_value_ or ""
        outer["To"] = self.to_ or ""
        if self.cc_ is not None:
            outer["Cc"] = self.cc_
        if self.bcc_ is not None:
            outer["Bcc"] = self.bcc_
        if self.organization_ is not None:
            outer["Organization"] = self.organization_
        if self.priority_ > 0:
            outer["Priority"] = str(self.priority_)
        outer["Subject"] = self.subject_ or ""
        outer.preamble = "You will not see this in a MIME-aware mail reader.\n"
        mime_type_ = "text/plain"

        if self.mime_type_ is not None:
            mime_type_ = self.mime_type_

        outer.add_header("Content-Type", mime_type_)

        outer.attach(MIMEText(self.body_ or "", mime_type_.split("/")[1], "utf-8"))

        step += 1
        self.sendStepNumber.emit(step)
        # Adjuntar logo
        if settings.SETTINGS.value("email/sendMailLogo", True):
            logo = settings.SETTINGS.value(
                "email/mailLogo", "%s/logo_mail.png" % application.PROJECT.tmpdir
            )
            if not QtCore.QFile.exists(logo):
                logo = "%s/logo.png" % application.PROJECT.tmpdir
                QtGui.QPixmap(utils_base.pixmap_from_mime_source("pineboo-logo.png")).save(
                    logo, "PNG"
                )

            fp_ = open(logo, "rb")
            logo_part = MIMEImage(fp_.read())
            fp_.close()

            logo_part.add_header("Content-ID", "<image>")
            outer.attach(logo_part)

        # Ficheros Adjuntos
        for att in self.attachments_:
            try:
                with open(att, "rb") as fil:
                    part = MIMEApplication(fil.read(), Name=basename(att))
                    part["Content-Disposition"] = 'attachment; filename="%s"' % basename(att)
                    outer.attach(part)
            except IOError:
                LOGGER.warning("Error al adjuntar el fichero %s." % att)
                return False

        # Envio mail
        composed = outer.as_string()

        step += 1
        self.sendStepNumber.emit(step)

        try:
            smtp = smtplib.SMTP("%s:%s" % (self.mail_server_ or "", self.port_ or 25))
            if self.connection_type_ == ConnectionType.TlsConnection:
                smtp.starttls()
                self.changeStatus("StartTTLS", State.StartTTLS)

            if self.user_ and self.password_:
                status_msg = "login."
                if self.auth_method_ == State.SendAuthLogin:
                    self.changeStatus(status_msg, State.SendAuthLogin)
                elif self.auth_method_ == State.SendAuthPlain:
                    self.changeStatus(status_msg, State.SendAuthPlain)

                smtp.login(self.user_, self.password_)

                self.changeStatus(
                    status_msg,
                    State.WaitingForAuthLogin
                    if self.auth_method_ == State.SendAuthLogin
                    else State.WaitingForAuthPlain,
                )

            smtp.sendmail(self.from_value_ or "", self.to_ or "", composed)
            self.changeStatus("Correo enviado", State.SendOk)
            smtp.quit()
            return True

        except smtplib.SMTPHeloError:
            status_msg = "El servidor no ha respondido correctamente al saludo."
            self.changeStatus(status_msg, State.ClientError)
            return False
        # except smtplib.SMTPNotSupportedError:
        #     status_msg = "El tipo de autenticación no está soportada por el servidor."
        #     self.changeStatus(status_msg, State.ClientError)
        #     return False
        except smtplib.SMTPConnectError:
            status_msg = "No se puede conectar al servidor SMTP."
            self.changeStatus(status_msg, State.ServerError)
            return False
        except smtplib.SMTPAuthenticationError:
            status_msg = "Error de autenticación SMTP."
            self.changeStatus(status_msg, State.ClientError)
            return False
        except smtplib.SMTPSenderRefused:
            status_msg = "Dirección de remitente rechazada."
            self.changeStatus(status_msg, State.ClientError)
            return False
        except smtplib.SMTPRecipientsRefused:
            status_msg = "Todas las direcciones de destinatarios se rechazaron."
            self.changeStatus(status_msg, State.ClientError)
            return False
        except smtplib.SMTPServerDisconnected:
            status_msg = "El servidor se desconecta inesperadamente."
            self.changeStatus(status_msg, State.ServerError)
            return False
        except smtplib.SMTPException:
            status_msg = "Error desconocido"
            self.changeStatus(status_msg, State.ClientError)
            return False
        except socket.gaierror:
            status_msg = (
                "Servidor SMTP no encontrado.Verifique el nombre de host de su servidor SMTP."
            )
            self.changeStatus(status_msg, State.SmtpError)
            return False
        except Exception as error:
            status_msg = "Error sending mail %s." % error
            return False

    def changeStatus(self, status_msg: str, state_code: int) -> None:
        """Change send mail status."""
        self.status_msg_ = status_msg
        self.state_code_ = state_code
        self.statusChanged.emit(self.status_msg_, self.state_code_)
        self.status.emit(self.status_msg_)



"""
Utility functions for QS files.
"""
import traceback
import re
import math
import sys
import threading
import json
import os

from PyQt6 import QtCore

from pineboolib.application import types, qsadictmodules
from pineboolib.core.utils import utils_base, logging

from pineboolib import application

from typing import (
    Any,
    Optional,
    Union,
    Match,
    List,
    Generator,
    Callable,
    Iterable,
    Dict,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from sqlalchemy.engine import (  # type: ignore [import] # noqa: F401, F821
        base,
    )  # pragma: no cover
    from pineboolib.interfaces import isession  # pragma: no cover

LOGGER = logging.get_logger(__name__)

TIMERS: List[QtCore.QTimer] = []


class Switch(object):
    """
    Switch emulation class.

    from: http://code.activestate.com/recipes/410692/
    This class provides the functionality we want. You only need to look at
    this if you want to know how this works. It only needs to be defined
    once, no need to muck around with its internals.
    """

    def __init__(self, value: Any):
        """Construct new witch from initial value."""
        self.value = value
        self.fall = False

    def __iter__(self) -> Generator:
        """Return the match method once, then stop."""
        yield self.match

    def match(self, *args: List[Any]) -> bool:
        """Indicate whether or not to enter a case suite."""
        if self.fall or not args:
            return True
        elif self.value in args:
            self.fall = True
            return True
        else:
            return False


class QsaRegExp(object):
    """
    Regexp emulation class.
    """

    result_: Optional[Match[str]]

    def __init__(self, str_re: str, is_global: bool = False):
        """Create new regexp."""
        self.str_re = str_re
        self.pattern = re.compile(self.str_re)
        self.is_global = is_global
        self.result_ = None

    def search(self, text: str) -> Optional[Match[str]]:
        """Return Match from search."""
        self.result_ = None
        if self.pattern is not None:
            self.result_ = self.pattern.search(text)
        return self.result_

    def replace(self, target: str, new_value: str) -> str:
        """Replace string using regex."""
        count = 1 if not self.is_global else 0
        return self.pattern.sub(new_value, target, count)

    def cap(self, i: int) -> Optional[str]:
        """Return regex group number "i"."""
        if self.result_ is None:
            return None

        try:
            return self.result_.group(i)
        except Exception:
            LOGGER.exception("Error calling cap(%s)" % i)
            return None

    def get_global(self) -> bool:
        """Return if regex is global."""
        return self.is_global

    def set_global(self, state: bool) -> None:
        """Set regex global flag."""
        self.is_global = state

    def exactMatch(self, value: str) -> Optional[Match[str]]:
        """Return exactMatch."""

        return self.pattern.fullmatch(value)

    global_ = property(get_global, set_global)


def reg_exp(str_re: str) -> QsaRegExp:
    """
    Return qsaRegexp object from search.

    @param strRE. Cadena de texto
    @return valor procesado
    """
    is_global = False
    if str_re[-2:] == "/g":
        str_re = str_re[:-2]
        is_global = True
    elif str_re[-1:] == "/":
        str_re = str_re[:-1]

    if str_re[:1] == "/":
        str_re = str_re[1:]

    return QsaRegExp(str_re, is_global)


class MathClass(object):
    """QSA Math emulation class."""

    def abs(self, value: Union[int, float]) -> Union[int, float]:
        """Get absolute value."""
        return math.fabs(value)

    def acos(self, value: Union[float, int]) -> float:
        """Return the arc cosine."""

        return math.acos(value)

    def cos(self, value: Union[float, int]) -> float:
        """Return the cosine of value."""

        return math.cos(value)

    def asin(self, value: Union[float, int]) -> float:
        """Return the arc sine of value."""

        return math.asin(value)

    def sin(self, value: Union[float, int]) -> float:
        """Return the sine of value."""

        return math.sin(value)

    def atan(self, value: Union[float, int]) -> float:
        """Return the arc tangent of value."""

        return math.atan(value)

    def atan2(self, value_y: Union[float, int], value_x: Union[float, int]) -> float:
        """Return the arc tangent."""

        return math.atan2(value_y, value_x)

    def tan(self, value: Union[float, int]) -> float:
        """Return the tangent of value."""

        return math.tan(value)

    def exp(self, value: Union[float, int]) -> float:
        """Return e raised to the power of value."""

        return math.exp(value)

    def ceil(self, value: float) -> int:
        """Round number to its ceiling."""
        return math.ceil(value)

    def floor(self, value: float) -> int:
        """Round number to its floor."""
        return math.floor(value)

    def log(self, value: Union[float, int]) -> float:
        """Return the logarithm of value to the given base."""

        return math.log(value)

    def random(self) -> float:
        """Return a pseudo-random floating point number between 0 and 1."""
        import random

        return random.random()

    def max(self, number1: Union[float, int], number2: Union[float, int]) -> Union[float, int]:
        """Return the largest of number1 and number2."""

        return max([number1, number2])

    def min(self, number1: Union[float, int], number2: Union[float, int]) -> Union[float, int]:
        """Return the smallest of number1 and number2."""

        return min([number1, number2])

    def pow(self, base_: float, exp: float) -> float:
        """Raise base to the power of exp."""
        return math.pow(base_, exp)

    def round(self, value_1: float, value_2: Optional[int] = None) -> float:
        """Round a number x to y decimal places."""
        return round(float(value_1), value_2)

    def sqrt(self, value: Union[float, int]) -> float:
        """Return the square root of the number passed in the parameter."""

        return math.sqrt(value)

    def _get_pi(self) -> float:
        """Return PI value."""

        return 3.141592653589793

    def _get_eulen(self) -> float:
        """Return eulers constant. The base for natural logarithms."""

        return 2.718281828459045

    def _get_ln2(self) -> float:
        """Return natural logarithm of 2."""

        return 0.6931471805599453

    def _get_ln10(self) -> float:
        """Return natural logarithm of 10."""

        return 2.302585092994046

    def _get_log2e(self) -> float:
        """Return base 2 logarithm of E."""

        return 1.44269504089

    def _get_log10e(self) -> float:
        """Return base 2 logarithm of E."""

        return 0.4342944819

    def _get_sqrt1_2(self) -> float:
        """Return square root of 1/2."""

        return 0.7071067811865476

    def _get_sqrt2(self) -> float:
        """Return square root of 2."""

        return 1.4142135623730951

    PI = property(_get_pi)
    E = property(_get_eulen)
    LN2 = property(_get_ln2)
    LN10 = property(_get_ln10)
    LOG2E = property(_get_log2e)
    LOG10E = property(_get_log10e)
    SQRT1_2 = property(_get_sqrt1_2)
    SQRT2 = property(_get_sqrt2)


def parse_float(value: Any) -> float:
    """
    Convert to float from almost any value.

    @param value. valor a convertir
    @return Valor tipo float, o parametro x , si no es convertible
    """
    ret = 0.00
    try:
        if isinstance(value, str) and value.find(":") > -1:
            # Convertimos a horas
            list_ = value.split(":")
            value = float(list_[0])  # Horas
            value += float(list_[1]) / 60  # Minutos a hora
            value += float(list_[2]) / 3600  # Segundos a hora

        if isinstance(value, str):
            try:
                return float(value)
            except Exception:
                value = value.replace(".", "")
                value = value.replace(",", ".")
                try:
                    return float(value)
                except Exception:
                    return float("nan")

        else:
            ret = 0.0 if value in (None, "") else float(value)

        if ret == int(ret):
            return int(ret)

        return ret
    except Exception:
        LOGGER.exception("parseFloat: Error converting %s to float" % value, stack_info=True)
        return float("nan")


def parse_string(obj: Any) -> str:
    """
    Convert to string almost any value.

    @param obj. valor a convertir
    @return str del objeto dado
    """
    return obj.toString() if hasattr(obj, "toString") else str(obj)


def parse_int(value: Union[float, int, str], base_: int = 10) -> int:
    """
    Convert to int almost any value.

    @param x. Value to cenvert
    @return integer value
    """
    ret_ = 0

    tmp_value = str(value)
    if tmp_value.find(".") > -1:
        tmp_value = tmp_value[0 : tmp_value.find(".")]

    if tmp_value.find(",") > -1:
        tmp_value = tmp_value[0 : tmp_value.find(",")]

    if value is not None:
        # x = float(x)
        ret_ = int(tmp_value, base_)
        # ret_ = int(str(x), base)

    return ret_


def length(obj: Any = "") -> int:
    """
    Get length of any object.

    @param obj, objeto a obtener longitud
    @return longitud del objeto
    """
    if obj is None:
        return 0

    if hasattr(obj, "length"):
        if isinstance(obj.length, int):
            return obj.length
        else:
            return obj.length()

    else:
        if isinstance(obj, dict) and "result" in obj.keys():
            return len(obj) - 1
        else:
            return len(obj)


def text(obj: Any) -> str:
    """
    Get text property from object.

    @param obj. Objeto a procesar
    @return Valor de text o text()
    """
    try:
        return obj.text()
    except Exception:
        return obj.text


def start_timer(time: int, fun: Callable) -> "QtCore.QTimer":
    """Create new timer that calls a function."""
    global TIMERS
    timer = QtCore.QTimer()
    timer.timeout.connect(fun)  # type: ignore [attr-defined] # noqa: F821
    timer.start(time)
    TIMERS.append(timer)
    return timer


def kill_timer(timer: Optional["QtCore.QTimer"] = None) -> None:
    """Stop a given timer."""
    global TIMERS
    if timer is not None:
        timer.stop()
        TIMERS.remove(timer)


def kill_timers() -> None:
    """Stop and deletes all timers that have been created with startTimer()."""
    global TIMERS
    for timer in TIMERS:
        timer.stop()

    TIMERS = []


def debug(txt: Union[bool, str, int, float]) -> None:
    """
    Debug for QSA messages.

    @param txt. Mensaje.
    """
    from pineboolib import application

    application.PROJECT.message_manager().send("debug", None, [utils_base.ustr(txt)])


def format_exc(exc: Optional[int] = None) -> str:
    """Format a traceback."""
    return traceback.format_exc(exc)


def is_nan(value: Any) -> bool:
    """
    Check if value is NaN.

    @param x. Valor numérico
    @return True o False
    """
    if value in [None, ""] or isinstance(value, bool):
        return True

    if isinstance(value, str) and value.find(":"):
        value = value.replace(":", "")
    try:
        value = float(value)
        return math.isnan(value)
    except ValueError:
        return True


def isnan(value: Any) -> bool:
    """Return if a number is NaN."""
    return is_nan(value)


def replace(source: str, search: Any, replace: str) -> str:
    """Replace for QSA where detects if "search" is a Regexp."""
    if isinstance(search, str):
        return source.replace(search, str(replace))
    else:
        return search.replace(source, replace)


def splice(*args: Any) -> Any:
    """Splice Iterables."""

    real_args = args[1:]
    array_ = args[0]
    if hasattr(array_, "splice"):
        array_.splice(real_args)
    else:
        if isinstance(array_, list):
            new_array_ = []

            if len(real_args) == 2:  # Delete
                pos_ini = real_args[0]
                length_ = real_args[1]
                for i in reversed(range(pos_ini, pos_ini + length_)):
                    array_.pop(i)

            elif len(real_args) > 2 and real_args[1] == 0:  # Insertion
                for value in reversed(real_args[2:]):
                    array_.insert(real_args[0], value)

            elif len(real_args) > 2 and real_args[1] > 0:  # Replacement
                pos_ini = real_args[0]
                replacement_size = real_args[1]
                new_values = real_args[2:]

                count_1 = 0
                count_2 = 0
                for old_value in array_:
                    if count_1 < pos_ini:
                        new_array_.append(old_value)
                    else:
                        if count_2 < replacement_size:
                            if count_2 == 0:
                                for new_value in new_values:
                                    new_array_.append(new_value)

                            count_2 += 1
                        else:
                            new_array_.append(old_value)

                    count_1 += 1
                array_ = new_array_


class Sort:
    """Sort Class."""

    _function: Optional[Callable] = None

    def __init__(self, function: Optional[Callable] = None):
        """Initialize function."""
        self._function = function

    def sort_(self, array_: Iterable) -> List:
        """Sort function."""

        new_array_: List = []
        if self._function is not None:
            for pos, value in enumerate(array_):
                found = False
                for new_pos, new_value in enumerate(list(new_array_)):
                    result = self._function(value, new_value)
                    # print("Comparando", value, new_value, result, "-->", new_array_)
                    if result == 0:
                        # print("Es igual (%s == %s)" % (value, new_value))
                        new_array_.append(value)
                        found = True
                        break
                    elif result == 1:
                        # print("Es mayor (%s > %s)" % (value, new_value))
                        continue

                    elif result == -1:
                        # print("Es menor (%s < %s)" % (value, new_value))
                        new_array_.insert(new_pos, value)
                        found = True
                        break

                if not found:
                    new_array_.append(value)
        else:
            new_array_ = sorted(array_)

        array_ = new_array_
        return new_array_


class NumberAttr:
    """Class Number_attr."""

    MIN_VALUE = -sys.maxsize - 1
    MAX_VALUE = sys.maxsize


def user_id() -> str:
    """Return user_id."""

    result = getattr(qsadictmodules.QSADictModules.from_project("sys").iface, "current_user")
    if not result:
        result = application.PROJECT.session_id()
    return result


def set_user_id(user_id: str) -> None:
    """Set user id."""

    qsadictmodules.from_project("sys").iface.current_user = user_id


def driver_session(conn_name: str = "default") -> "isession.PinebooSession":
    """Return driver session."""

    return application.PROJECT.conn_manager.useConn(conn_name).driver().session()


def session(conn_name: str = "default", legacy: bool = False) -> "isession.PinebooSession":
    """Return session connection."""

    return (
        application.PROJECT.conn_manager.useConn(conn_name).session()
        if legacy
        else driver_session(conn_name)
    )


def thread_session_new(conn_name: str = "default") -> "isession.PinebooSession":
    """Return thread session new."""

    return session(conn_name, True)


def thread_session_current(conn_name: str = "default") -> Optional["isession.PinebooSession"]:
    """Return session current."""

    return application.PROJECT.conn_manager.useConn(conn_name)._session_legacy


def is_valid_session(session: "isession.PinebooSession", raise_error: bool = False) -> bool:
    """Return if a session is valid."""

    return application.PROJECT.conn_manager.is_valid_session(session, raise_error)


def thread_session_free(conn_name: str = "default") -> None:
    """Close and delete current thread session."""

    session = application.PROJECT.conn_manager.useConn(conn_name)._session_legacy
    application.PROJECT.conn_manager.useConn(conn_name)._session_legacy = None
    if session is not None:
        session.close()


def thread() -> int:
    """Return thread id."""

    return threading.current_thread().ident or -1


def session_atomic(conn_name: str = "default") -> Optional["isession.PinebooSession"]:
    """Return atomic_session."""

    return application.PROJECT.conn_manager.useConn(conn_name)._session_atomic


def ws_channel_send(msg: Any = "", group_name: str = "") -> None:
    """Send message to websocket channel."""

    if application.USE_WEBSOCKET_CHANNEL:
        from asgiref.sync import async_to_sync  # type: ignore [import] # noqa: F723
        from channels.layers import get_channel_layer  # type: ignore [import] # noqa: F723

        json = {"type": "send.msg", "content": msg}

        channel_layer = get_channel_layer()
        user_id = application.PROJECT.session_id()
        if group_name:
            async_to_sync(channel_layer.group_send)(group_name, json)
        else:
            async_to_sync(channel_layer.send)(user_id, json)


def ws_channel_send_type(json: Dict, group_name: str = "") -> None:
    """Send message to websocket channel."""

    if application.USE_WEBSOCKET_CHANNEL:
        from asgiref.sync import async_to_sync  # type: ignore [import] # noqa: F723
        from channels.layers import get_channel_layer  # type: ignore [import] # noqa: F723

        channel_layer = get_channel_layer()
        user_id = application.PROJECT.session_id()
        if group_name:
            async_to_sync(channel_layer.group_send)(group_name, json)
        else:
            async_to_sync(channel_layer.send)(user_id, json)


def typeof_(obj: Any) -> str:
    """Return type name froma n object."""

    result = "unknown"

    if isinstance(obj, str):
        result = "string"
    elif isinstance(obj, bool):
        result = "boolean"
    elif isinstance(obj, (int, float)):
        result = "number"
    elif isinstance(obj, (dict, list, types.Array)):
        result = "object"
    elif hasattr(obj, "__call__"):
        result = "function"

    return result


def _super(class_name: str, obj: Callable) -> "super":
    """Super class."""

    for classes in obj.__class__.__mro__:
        if classes.__name__ == class_name:
            return super(classes, obj)

    raise Exception("Superclass %s not found." % class_name)


def pool_status(conn_name: str = "main_conn") -> str:
    """Return pool status used for the conn_name."""

    return application.PROJECT.conn_manager.pool_status(conn_name)


def memory_status() -> None:
    """Return memory status."""

    from pineboolib.fllegacy import flutil

    time_stamp = flutil.FLUtil().timestamp()
    file_path = os.path.join(application.PROJECT.tmpdir, "%s.txt" % time_stamp)

    file_ = open(file_path, "w", encoding="UTF-8")

    try:
        from pympler import muppy, summary  # type: ignore [import] # noqa: F821

        all_objects = muppy.get_objects()

        file_.write("TYPES,OBJECTS,MEMORY SIZE")
        sum_memory = 0
        sum_objects = 0
        for item in summary.summarize(all_objects):
            file_.write("\n%s" % item)
            sum_objects += item[1]
            sum_memory += item[2]

        file_.write("\nOBJECTS IN MEMORY  : %s" % len(all_objects))
        file_.write("\nTOTAL MEMORY USAGE : %s" % sum_memory)
        file_.close()
        LOGGER.warning(
            "MEMORY_STATUS: File %s is created with debug. Usage(%s)", file_path, sum_memory
        )

    except ImportError:
        LOGGER.warning("need install 'pympler' module first.")


def qt_translate_noop(string: str, path: str, mod: str) -> str:
    """Return a translation."""

    return utils_base.qt_translate_noop(string, path, mod)


def require(name: str) -> Any:
    """Return require."""

    return qsadictmodules.from_project("formImport").from_(name)


def to_json(text: str) -> Union[Dict, List]:
    """Return json object from str."""
    text = text.replace("'", '"')
    return json.loads(text)



"""Test Eneboo module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import application

from pineboolib.core import settings


class TestEnebooGUI(unittest.TestCase):
    """Tes EnebooGUI class."""

    prev_main_window_name: str

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        settings.CONFIG.set_value("application/isDebuggerMode", True)
        settings.CONFIG.set_value("application/dbadmin_enabled", True)
        cls.prev_main_window_name = settings.CONFIG.value(
            "ebcomportamiento/main_form_name", "eneboo"
        )
        settings.CONFIG.set_value("ebcomportamiento/main_form_name", "eneboo_mdi")

        init_testing()

    def test_initialize(self) -> None:
        """Test GUI initialize."""
        # from pineboolib.qsa import qsa
        from pineboolib.plugins.mainform.eneboo_mdi import eneboo_mdi

        # import os

        # application.PROJECT.main_form = eneboo_mdi
        # eneboo_mdi.mainWindow = eneboo_mdi.MainForm()
        # eneboo_mdi.mainWindow.initScript()

        application.PROJECT.main_window = eneboo_mdi.MainForm()
        self.assertTrue(application.PROJECT.main_window)
        application.PROJECT.main_window.initScript()

        # qsa_sys = qsa.sys
        # path = fixture_path("principal.eneboopkg")
        # self.assertTrue(os.path.exists(path))
        # qsa_sys.loadModules(path, False)
        # application.PROJECT.main_window = application.PROJECT.main_form.mainWindow  # type: ignore
        self.assertTrue(application.PROJECT.main_window)
        if application.PROJECT.main_window:
            application.PROJECT.main_window.initToolBar()
            application.PROJECT.main_window.windowMenuAboutToShow()
            application.PROJECT.main_window.show()
            application.PROJECT.main_window.activateModule("sys")
            # if application.PROJECT.main_window._p_work_space is not None:
            #    for window in application.PROJECT.main_window._p_work_space.subWindowList():

            self.assertFalse(application.PROJECT.main_window.existFormInMDI("flusers"))
            application.PROJECT.actions["flusers"].openDefaultForm()
            application.PROJECT.main_window.windowMenuAboutToShow()
            application.PROJECT.main_window.windowMenuActivated(0)
            self.assertTrue(application.PROJECT.main_window.existFormInMDI("flusers"))
            application.PROJECT.main_window.writeState()
            application.PROJECT.main_window.writeStateModule()
            application.PROJECT.main_window.toggleToolBar(True)
            application.PROJECT.main_window.toggleStatusBar(True)
            application.PROJECT.main_window.windowClose()

            application.PROJECT.main_window.generalExit(False)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure this class is finished correctly."""

        del application.PROJECT.main_window

        settings.CONFIG.set_value("application/isDebuggerMode", False)
        settings.CONFIG.set_value("application/dbadmin_enabled", False)
        settings.CONFIG.set_value("ebcomportamiento/main_form_name", cls.prev_main_window_name)

        finish_testing()



"""Flmysql_innodb module."""
from pineboolib.plugins.sql.flmysql_myisam import FLMYSQL_MYISAM


class FLMYSQL_INNODB(FLMYSQL_MYISAM):  # pylint: disable=invalid-name
    """FLMYSQL_INNODB class."""

    def __init__(self):
        """Inicialize."""

        super().__init__()
        self.name_ = "FLMYSQL_INNODB"
        self.alias_ = "MySQL InnoDB (MYSQLDB)"
        self._no_inno_db = False
        self._default_charset = "DEFAULT CHARACTER SET = UTF8MB4 COLLATE = UTF8MB4_BIN"



"""
Tests for application.types module.
"""

import unittest
import os
from pineboolib.loader.main import init_cli
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.core import settings
from pineboolib.application import types

import datetime


init_cli()  # FIXME: This should be avoided


class TestBoolean(unittest.TestCase):
    """Test booleans."""

    def test_true(self) -> None:
        """Test for true."""
        self.assertEqual(types.boolean(1), True)
        self.assertEqual(types.boolean("True"), True)
        self.assertEqual(types.boolean("Yes"), True)
        self.assertEqual(types.boolean(0.8), True)
        self.assertEqual(types.boolean(True), True)

    def test_false(self) -> None:
        """Test for false."""
        self.assertEqual(types.boolean(0), False)
        self.assertEqual(types.boolean("False"), False)
        self.assertEqual(types.boolean("No"), False)
        self.assertEqual(types.boolean(False), False)


class TestQString(unittest.TestCase):
    """Test QString."""

    def test_basic(self) -> None:
        """Basic testing."""
        text = types.QString("hello world")
        self.assertEqual(text, "hello world")
        self.assertEqual(text.mid(5), text[5:])
        self.assertEqual(text.mid(5, 2), text[5:7])


class TestFunction(unittest.TestCase):
    """Test function. Parses QSA into Python."""

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()

    def test_basic(self) -> None:
        """Basic testing."""
        source = "return x + 1"
        fun_ = types.function("x", source)
        self.assertEqual(fun_(1), 2)

    def test_extended(self) -> None:
        """Extended testing."""

        qsa_src = [
            "/* Parámetros del objeto o que pueden ser usados en la fórmula:",
            "* o.DS: Días de servicio",
            "* o.T1: Ventas en el trimestre T1",
            "* o.T2: Ventas en el trimestre T2",
            "* o.T3: Ventas en el trimestre T3",
            "* o.T4: Ventas en el trimestre T4",
            "* o.SA: Stock actual",
            "* o.RS: Stock reservado",
            "* o.PR: Stock pendiente de recibir",
            "* o.SS: Stock de seguridad",
            "*/",
            "o = arguments[0]; /// NO BORRES ESTAS LINEAS",
            "for (d in o) {",
            "o[d] = parseFloat(o[d]);",
            "}",
            "//////////////////////////////////////////////////",
            "consumoMensual = (parseFloat(o.T1) + parseFloat(o.T2)) / 2;",
            "numeroMeses = o.DS/30;",
            "aPedir = consumoMensual * numeroMeses - parseFloat(o.SA) + parseFloat(o.RS) - parseFloat(o.PR);",
            "if (aPedir < 0) {",
            "aPedir=0;",
            "}",
            "aPedir = Math.ceil(aPedir);",
            "return aPedir;",
        ]

        fun_ = types.function("arguments", "\n".join(qsa_src))
        data = types.Array()
        data.T1 = 100
        data.T2 = 200
        data.DS = 6

        result = fun_([data])
        self.assertEqual(result, 30)

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()


class TestObject(unittest.TestCase):
    """Test object."""

    def test_basic1(self) -> None:
        """Basic testing."""
        object_ = types.object_()
        object_.prop1 = 1
        object_.prop2 = 2
        self.assertEqual(object_.prop1, object_["prop1"])

    def test_basic2(self) -> None:
        """Basic testing."""
        object_ = types.object_({"prop1": 1})
        self.assertEqual(object_.prop1, object_["prop1"])


class TestArray(unittest.TestCase):
    """Test Array class."""

    def test_basic1(self) -> None:
        """Basic testing."""
        array_ = types.Array()
        array_.value = 1
        self.assertEqual(array_.value, array_["value"])

    def test_basic2(self) -> None:
        """Basic testing."""
        test_arr = [0, 1, 2, 3, 4]
        array_ = types.Array(test_arr)
        array_b = types.Array(test_arr)
        self.assertEqual(array_[3], 3)
        self.assertEqual(list(array_._dict.values()), test_arr)
        self.assertEqual(len(array_), len(test_arr))
        self.assertEqual(array_, test_arr)
        self.assertEqual(array_[3], array_b[3])
        self.assertNotEqual(array_[3], array_b[0])

        test_arr = [3, 4, 2, 1, 0]
        array_ = types.Array(test_arr)
        self.assertEqual(list(array_._dict.values()), test_arr)
        array_.append(10)
        self.assertEqual(array_[5], 10)

    def test_basic3(self) -> None:
        """Basic Testing."""
        test_arr = {"key_0": "item_0", "key_1": "item_1", "key_2": "item_2"}
        array_ = types.Array(test_arr)
        self.assertEqual(array_["key_0"], "item_0")
        self.assertEqual(array_.key_1, array_["key_1"])
        self.assertEqual(array_.length(), 3)
        self.assertEqual(array_[2], "item_2")
        self.assertEqual(list(array_._dict.values()), ["item_0", "item_1", "item_2"])

    def test_repr(self) -> None:
        """Test repr method."""
        test_arr = [3, 4, 5, 6, 7]
        array_ = types.Array(test_arr)
        self.assertEqual(repr(array_), "<Array %r>" % test_arr)

    def test_iter(self) -> None:
        """Test iterating arrays."""

        test_arr = [3, 4, 5, 6, 7]
        array_ = types.Array(test_arr)
        array_2 = [x for x in array_]
        self.assertEqual(test_arr, array_2)

        test_arr = [8, 7, 6, 4, 2]
        array_ = types.Array(test_arr)
        array_2 = [x for x in array_]
        self.assertEqual(test_arr, array_2)

    def test_splice(self) -> None:
        """Test splice."""

        test_arr = [3, 4, 5, 6, 7]
        array_ = types.Array(test_arr)
        array_.splice(1, 2)  # Delete
        self.assertEqual(str(array_), str(types.Array([4, 5])))
        array_2 = types.Array(test_arr)
        array_2.splice(2, 0, 9, 10)  # Insertion
        self.assertEqual(str(array_2), str(types.Array([3, 4, 5, 9, 10, 6, 7])))
        array_3 = types.Array(test_arr)
        array_3.splice(2, 1, 9, 10)  # Replace
        self.assertEqual(str(array_3), str(types.Array([3, 4, 9, 10, 6, 7])))

    def test_concat(self) -> None:
        """Test concat."""
        rules = [
            {"idregla": "pedidoscli", "grupo": "pedidoscli", "descripcion": "Pedidos de cliente"},
            {
                "idregla": "pedidoscli/get",
                "grupo": "pedidoscli",
                "descripcion": "Puede recibir pedidos de cliente",
            },
            {
                "idregla": "pedidoscli/post",
                "grupo": "pedidoscli",
                "descripcion": "Puede crear pedidos de cliente",
            },
            {
                "idregla": "pedidoscli/patch",
                "grupo": "pedidoscli",
                "descripcion": "Puede modificar pedidos de cliente",
            },
            {
                "idregla": "pedidoscli/delete",
                "grupo": "pedidoscli",
                "descripcion": "Puede eliminar pedidos de cliente",
            },
            {
                "idregla": "pedidoscli/accion1",
                "grupo": "pedidoscli",
                "descripcion": "Puede ejecutar la accion 1 de pedidos de cliente. Puede ejecutar la accion 1 de pedidos de cliente.",
            },
            {
                "idregla": "pedidoscli/accion2",
                "grupo": "pedidoscli",
                "descripcion": "Puede ejecutar la accion 2 de pedidos de cliente",
            },
        ]
        self.assertEqual(self.getRules(), rules)

    def getRules(self):
        from pineboolib.qsa import qsa

        result = qsa.Array().concat(
            self.getCrudRules("pedidoscli", "Pedidos de cliente"),
            qsa.Array(
                [
                    qsa.AttributeDict(
                        {
                            "idregla": ("pedidoscli/accion1"),
                            "grupo": ("pedidoscli"),
                            "descripcion": (
                                "Puede ejecutar la accion 1 de pedidos de cliente. Puede ejecutar la accion 1 de pedidos de cliente."
                            ),
                        }
                    ),
                    qsa.AttributeDict(
                        {
                            "idregla": ("pedidoscli/accion2"),
                            "grupo": ("pedidoscli"),
                            "descripcion": ("Puede ejecutar la accion 2 de pedidos de cliente"),
                        }
                    ),
                ]
            ),
        )
        return result

    def getCrudRules(self, grupo, descripcion):
        from pineboolib.qsa import qsa

        crudRules = qsa.Array(
            [
                qsa.AttributeDict({"id": ("get"), "desc": ("Puede recibir")}),
                qsa.AttributeDict({"id": ("post"), "desc": ("Puede crear")}),
                qsa.AttributeDict({"id": ("patch"), "desc": ("Puede modificar")}),
                qsa.AttributeDict({"id": ("delete"), "desc": ("Puede eliminar")}),
            ]
        )
        rules = qsa.Array(
            [
                qsa.AttributeDict(
                    {"idregla": (grupo), "grupo": (grupo), "descripcion": (descripcion)}
                )
            ]
        )

        i = 0
        while_pass = True
        while i < qsa.length(crudRules):
            if not while_pass:
                i += 1
                while_pass = True
                continue
            while_pass = False
            rules.append(
                qsa.AttributeDict(
                    {
                        "idregla": (qsa.ustr(grupo, "/", crudRules[i].id)),
                        "grupo": (grupo),
                        "descripcion": (qsa.ustr(crudRules[i].desc, " ", descripcion.lower())),
                    }
                )
            )
            i += 1
            while_pass = True
            try:
                i < qsa.length(crudRules)
            except Exception:
                break
        return rules


class TestDate(unittest.TestCase):
    """Test Date class."""

    # FIXME: Complete unit tests
    def test_basic1(self) -> None:
        """Basic testing."""
        date_ = types.Date("2001-02-25")
        self.assertEqual(date_.getDay(), 25)
        self.assertEqual(date_.getMonth(), 2)
        self.assertEqual(date_.getYear(), 2001)

    def test_basic2(self) -> None:
        today = datetime.date.today()
        date_ = types.Date(today)
        self.assertEqual(date_.getDay(), today.day)
        self.assertEqual(date_.getMonth(), today.month)
        self.assertEqual(date_.getYear(), today.year)

        other = datetime.datetime.strptime("2001-01-01", "%Y-%m-%d")

        date_ = types.Date(other)
        self.assertEqual(date_.getDay(), other.day)
        self.assertEqual(date_.getMonth(), other.month)
        self.assertEqual(date_.getYear(), other.year)


class TestString(unittest.TestCase):
    """TestString class."""

    # FIXME: Complete unit tests
    def test_fromCharCode(self) -> None:
        """Test fromCharCode."""
        temp: str = types.String.fromCharCode(13, 10)
        self.assertEqual(temp, "\r\n")
        temp2: str = types.String.fromCharCode()
        self.assertEqual(temp2, "")


class TestFile(unittest.TestCase):
    """Test File class."""

    def test_write_read_values_1(self) -> None:
        """Check that you read the same as you write."""

        temporal = "%s%s" % (
            settings.CONFIG.value("ebcomportamiento/temp_dir"),
            "/test_types_file.txt",
        )
        contenido = 'QT_TRANSLATE_NOOP("MetaData","Código")'
        contenido_3 = 'QT_TRANSLATE_NOOP("MetaData","Código")'
        types.File(temporal).write(contenido)
        contenido_2 = types.File(temporal).read()
        self.assertEqual(contenido, contenido_2)
        os.remove(temporal)
        types.File(temporal).write(contenido_3)
        contenido_4 = types.File(temporal).read()
        self.assertEqual(contenido_3, contenido_4)
        os.remove(temporal)

    def test_write_read_values_2(self) -> None:
        """Check that you read the same as you write."""

        temporal = "%s%s" % (
            settings.CONFIG.value("ebcomportamiento/temp_dir"),
            "/test_types_file_static.txt",
        )
        contenido = 'QT_TRANSLATE_NOOP("MetaData","Código")'
        types.FileStatic.write(temporal, contenido)
        contenido_2 = types.FileStatic.read(temporal)
        self.assertEqual(contenido, contenido_2)
        os.remove(temporal)

    def test_write_read_bytes_1(self) -> None:
        """Check that you read the same as you write."""

        temporal = "%s%s" % (
            settings.CONFIG.value("ebcomportamiento/temp_dir"),
            "/test_types_file_bytes.txt",
        )
        contenido = "Texto escrito en bytes\n".encode("utf-8")
        types.File(temporal).write(contenido)
        contenido_2 = types.File(temporal).read()
        self.assertEqual(contenido, contenido_2.encode("utf-8"))
        os.remove(temporal)

    def test_write_read_byte_1(self) -> None:
        """Check that you read the same as you write."""

        temporal = "%s%s" % (
            settings.CONFIG.value("ebcomportamiento/temp_dir"),
            "/test_types_file_bytes.txt",
        )
        contenido = "Texto\n".encode("utf-8")
        types.File(temporal).write(contenido)
        contenido_2 = types.File(temporal).read()
        self.assertEqual(contenido, contenido_2.encode("utf-8"))
        os.remove(temporal)

    def test_write_read_line_1(self) -> None:
        """Check that you read the same as you write."""

        temporal = "%s%s" % (
            settings.CONFIG.value("ebcomportamiento/temp_dir"),
            "/test_types_file_lines.txt",
        )
        contenido = "Esta es la linea"
        types.File(temporal).writeLine("%s 1" % contenido)
        types.File(temporal).writeLine("%s 2" % contenido, 4)
        file_read = types.File(temporal)
        linea_1 = file_read.readLine()
        self.assertEqual("%s 1\n" % contenido, linea_1)
        linea_2 = file_read.readLine()
        self.assertEqual("%s" % contenido[0:4], linea_2)
        os.remove(temporal)

    def test_full_name_and_readable(self) -> None:
        """Check fullName"""

        temporal = "%s%s" % (
            settings.CONFIG.value("ebcomportamiento/temp_dir"),
            "/test_types_file_full_name.txt",
        )
        contenido = 'QT_TRANSLATE_NOOP("MetaData","Código")'
        file_ = types.File(temporal)
        file_.write(contenido)
        self.assertEqual(file_.fullName, temporal)
        self.assertTrue(file_.readable())

    def test_last_modified(self) -> None:
        """Test lastModified."""

        temporal = "%s%s" % (
            settings.CONFIG.value("ebcomportamiento/temp_dir"),
            "/test_last_modified.txt",
        )
        contenido = 'QT_TRANSLATE_NOOP("MetaData","Código")'
        file_ = types.File(temporal)
        file_.write(contenido)
        file_.close()
        self.assertNotEqual(file_.lastModified(), "")

    def test_properties(self) -> None:
        temporal = "%s%s" % (
            settings.CONFIG.value("ebcomportamiento/temp_dir"),
            "/test_last_modified.txt",
        )
        file_ = types.File(temporal)
        self.assertEqual(file_.path, settings.CONFIG.value("ebcomportamiento/temp_dir"))
        self.assertEqual(file_.fullName, temporal)
        self.assertEqual(file_.extension, ".txt")
        self.assertEqual(file_.baseName, "test_last_modified")
        self.assertTrue(file_.exists)
        self.assertEqual(file_.size, 38)


class TestDir(unittest.TestCase):
    """TestDir class."""

    def test_current(self) -> None:
        """Check Dir."""

        self.assertEqual(os.curdir, types.Dir().current)
        self.assertEqual(os.curdir, types.DirStatic.current)

    def test_mkdir_rmdir(self) -> None:
        """Test mkdir and rmdir."""

        tmp_dir = settings.CONFIG.value("ebcomportamiento/temp_dir")
        my_dir = types.Dir(tmp_dir)
        my_dir.mkdir("test")
        self.assertTrue(os.path.exists("%s/test" % tmp_dir))
        my_dir.rmdirs("test")
        self.assertFalse(os.path.exists("%s/test" % tmp_dir))

    def test_change_dir(self) -> None:
        """Test change dir."""

        tmp_dir = settings.CONFIG.value("ebcomportamiento/temp_dir")
        my_dir = types.Dir(tmp_dir)
        original_dir = my_dir.current
        # my_dir.mkdir("test_change_dir")
        # my_dir.cd("%s/test_change_dir" % tmp_dir)
        my_dir.cd(original_dir)
        self.assertEqual(my_dir.current, original_dir)
        my_dir.cdUp()
        # self.assertEqual(os.path.realpath(my_dir.current), tmp_dir)
        # my_dir.rmdirs("test_change_dir")
        my_dir.cd(original_dir)



"""Qtextedit module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets  # type: ignore[import]
from pineboolib.core import decorators
from typing import Optional


class QTextEdit(QtWidgets.QTextEdit):
    """QTextEdit class."""

    LogText: int = 0
    RichText: int = 1

    def __init__(self, parent: Optional["QtWidgets.QWidget"] = None) -> None:
        """Inicialize."""
        super().__init__(parent)
        self.LogText = 0  # pylint: disable=invalid-name
        self.RichText = 1  # pylint: disable=invalid-name

    def setText(self, text: Optional[str]) -> None:
        """Set text."""

        super(QTextEdit, self).setText(text)  # type: ignore [arg-type]
        # if not project.DGI.localDesktop():
        #    project.DGI._par.addQueque("%s_setText" % self._parent.objectName(), text)

    def getText(self) -> str:
        """Return text."""
        return super().toPlainText()

    @decorators.not_implemented_warn
    def getTextFormat(self) -> int:
        """Return text format."""
        return -1

    @decorators.incomplete
    def setTextFormat(self, value: int) -> None:
        """Set text format."""
        if value == 0:  # LogText
            self.setReadOnly(True)
            self.setAcceptRichText(False)
        elif value == 1:
            self.setReadOnly(False)
            self.setAcceptRichText(True)

    def setShown(self, value: bool) -> None:
        """Set visible."""
        self.setVisible(value)
        # if value:
        #    super().show()
        # else:
        #    super().hide()

    def getPlainText(self) -> str:
        """Return text in plain text format."""
        return super(QTextEdit, self).toPlainText()

    def setAutoFormatting(self, value=QtWidgets.QTextEdit.AutoFormattingFlag.AutoAll) -> None:
        """Set auto formating mode."""

        super().setAutoFormatting(QtWidgets.QTextEdit.AutoFormattingFlag.AutoAll)

    textFormat = property(getTextFormat, setTextFormat)
    text = property(getText, setText)
    PlainText = property(getPlainText, setText)



"""Converter to and from action_."""

from pineboolib.core.utils import logging

from typing import Union, cast, TYPE_CHECKING
from pineboolib import application
from pineboolib.application.metadata import pnaction

if TYPE_CHECKING:
    from pineboolib.application import xmlaction  # pragma: no cover


LOGGER = logging.get_logger(__name__)


def convert_from_flaction(action: "pnaction.PNAction") -> "xmlaction.XMLAction":
    """
    Convert a PNAction to XMLAction.

    @param action. action_ object.
    @return XMLAction object.
    """

    if action.name() not in application.PROJECT.actions.keys():
        raise KeyError("Action %s not loaded in current project" % action.name())
    return cast("xmlaction.XMLAction", application.PROJECT.actions[action.name()])


def convert_to_flaction(action: Union[str, "xmlaction.XMLAction"]) -> "pnaction.PNAction":
    """
    Convert a XMLAction to action_.

    @param action. XMLAction object.
    @return PNAction object.
    """

    action_name = action if isinstance(action, str) else action._name

    if application.PROJECT.conn_manager is None:
        raise Exception("Project is not connected yet")

    # LOGGER.trace("convert2action: Load action from db manager")

    cached_actions = application.PROJECT.conn_manager.manager()._cache_action
    if action_name in cached_actions.keys():
        return cached_actions[action_name]
    else:
        action_: "pnaction.PNAction" = pnaction.PNAction(action_name)
        if action_name in application.PROJECT.actions.keys():
            xml_action = application.PROJECT.actions[action_name]
            action_.setName(xml_action._name)
            action_.setTable(xml_action._table)
            action_.setForm(xml_action._master_form)
            action_.setFormRecord(xml_action._record_form)
            action_.setScriptForm(xml_action._master_script)
            action_.setScriptFormRecord(xml_action._record_script)
            action_.setDescription(xml_action._description)
            action_.setCaption(xml_action._alias or xml_action._caption)
            action_.setClass_(xml_action._class_script)

        cached_actions[action_name] = action_
        # LOGGER.trace("convert2FLAction: done")

    return action_



"""QDir module."""
# -*- coding: utf-8 -*-

from PyQt6 import QtCore  # type: ignore[import]


class QDir(QtCore.QDir):
    """QDir class."""

    def absPath(self):
        """Return absolute path."""
        return self.absolutePath()



"""Fldoublevalidator module."""
# -*- coding: utf-8 -*-


from PyQt6 import QtGui  # type: ignore[import]

from pineboolib import application

from typing import Tuple, Union, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from PyQt6 import QtWidgets  # pragma: no cover


class FLDoubleValidator(QtGui.QDoubleValidator):
    """FLDoubleValiator class."""

    _formatting: bool

    def __init__(
        self,
        max_value: Union[int, float],
        min_value: Union[int, float],
        part_decimal: int,
        parent: "QtWidgets.QWidget",
    ) -> None:
        """Inicialize."""
        super().__init__(max_value, min_value, part_decimal, parent)
        # 1 inferior
        # 2 superior
        # 3 partDecimal
        # 4 editor
        self.setNotation(self.Notation.StandardNotation)
        self._formatting = False

    def validate(
        self, input_: Optional[str], pos_cursor: int
    ) -> Tuple["QtGui.QValidator.State", str, int]:
        """Return if a value is valid."""
        value_in = input_

        if value_in is None or self._formatting:
            return (self.State.Acceptable, value_in, pos_cursor)  # type: ignore [return-value]

        # pos_cursor= len(value_in)
        state = super().validate(value_in, pos_cursor)
        # 0 Invalid
        # 1 Intermediate
        # 2 Acceptable
        ret_2 = state[2]

        if state[0] in (self.State.Invalid, self.State.Intermediate) and len(value_in) > 0:
            state_ = value_in[1:]
            if (
                value_in[0] == "-"
                and super().validate(state_, pos_cursor)[0] == self.State.Acceptable
                or state_ == ""
            ):
                ret_0 = self.State.Acceptable
            else:
                ret_0 = self.State.Invalid
        else:
            ret_0 = self.State.Acceptable

        ret_1 = state[1]

        if application.PROJECT.aq_app.commaSeparator() == "," and ret_1.endswith("."):
            ret_1 = ret_1[0 : len(ret_1) - 1] + ","

        if len(ret_1) == 1 and ret_1 not in (
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            ",",
            ".",
        ):
            ret_0 = self.State.Invalid
            ret_1 = ""
            ret_2 = 0

        return (ret_0, ret_1, ret_2)



"""Test_settings module."""

from PyQt6 import QtWidgets  # type: ignore[import]

import unittest
from pineboolib.loader.main import init_testing, finish_testing

from pineboolib.core import settings


class TestSettings(unittest.TestCase):
    """TestSettings Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_raises(self) -> None:
        """Test errors."""

        setting = settings.PinebooSettings("TEST")

        with self.assertRaises(Exception):
            sett = setting._value(  # type: ignore [arg-type] # noqa : F821
                QtWidgets.QWidget(), None  # type: ignore [arg-type] # noqa : F821
            )

        with self.assertRaises(TypeError):
            sett = setting.load_value(  # type: ignore [arg-type] # noqa : F821
                QtWidgets.QSizePolicy  # type: ignore [arg-type] # noqa : F821
            )

        with self.assertRaises(TypeError):
            sett = setting.load_value(QtWidgets.QLabel())  # type: ignore [arg-type] # noqa : F821

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Flfiles_dir module."""

import os
import hashlib
import pathlib
from typing import List


from PyQt6 import QtXml  # type: ignore[import]
from pineboolib.core.utils import logging, utils_base


LOGGER = logging.get_logger(__name__)


class FlFiles(object):
    """FlFiles class."""

    _root_dir: str
    _areas: List[List[str]]
    _modules: List[List[str]]
    _files: List[List[str]]

    def __init__(self, folder: str = "") -> None:
        """Initialize."""

        self._root_dir = folder
        self._areas = []
        self._modules = []
        self._files = []
        if os.path.exists(self._root_dir):
            self.build_data()
        else:
            LOGGER.warning("FLFILES_FOLDER: folder %s not found", self._root_dir)

    def areas(self) -> List[List[str]]:
        """Return areas info."""

        return self._areas

    def modules(self) -> List[List[str]]:
        """Return modules info."""

        return self._modules

    def files(self) -> List[List[str]]:
        """Return files info."""

        return self._files

    def build_data(self) -> None:
        """Build data from a folder."""

        for root, subdirs, files in os.walk(self._root_dir):
            module_found = None
            for file_name in files:
                if file_name.endswith(".mod"):
                    module_found = file_name
                    break

            if module_found:
                self.process_module(file_name, root, subdirs, files)

    def process_module(
        self, module_file: str, root_folder: str, subdirs: List[str], files: List[str]
    ) -> None:
        """Process a module folder."""

        nombre_fichero = os.path.join(root_folder, module_file)
        # print("Buscando ...", nombre_fichero)
        try:
            fichero = open(nombre_fichero, "r", encoding="iso-8859-15")
            datos_module = fichero.read()
            fichero.close()
        except Exception as error:
            LOGGER.error("Error processing %s:%s", nombre_fichero, str(error))
            return
        xml_module = QtXml.QDomDocument()
        descripcion_modulo: str
        if xml_module.setContent(datos_module):
            node_module = xml_module.namedItem("MODULE")
            modulo = node_module.namedItem("name").toElement().text()
            descripcion_modulo = node_module.namedItem("alias").toElement().text()
            area = node_module.namedItem("area").toElement().text()
            descripcion_area = node_module.namedItem("areaname").toElement().text()
            version = node_module.namedItem("version").toElement().text()
            nombre_icono = node_module.namedItem("icon").toElement().text()
            # if node_module.namedItem(u"flversion"):
            #    versionMinimaFL = node_module.namedItem(u"flversion").toElement().text()
            # if node_module.namedItem(u"dependencies") is not None:
            #    node_depend = xml_module.elementsByTagName(u"dependency")
            #    i = 0
            #    while i < len(node_depend):
            #        dependencias[i] = node_depend.item(i).toElement().text()
            #        i += 1

        descripcion_modulo = utils_base.qt_translate_noop(descripcion_modulo, root_folder, modulo)
        descripcion_area = utils_base.qt_translate_noop(descripcion_area, root_folder, modulo)
        datos_icono: str = ""
        if os.path.exists(os.path.join(root_folder, nombre_icono)):
            try:
                fichero_icono = open(
                    os.path.join(root_folder, nombre_icono), "r", encoding="ISO-8859-15"
                )
                datos_icono = fichero_icono.read()
                fichero_icono.close()
            except Exception as error:
                LOGGER.error("Error processing %s:%s", nombre_icono, str(error))
                return

        if area not in [idarea for idarea, descripcion_area in self._areas]:
            self._areas.append([area, descripcion_area])

        if modulo not in [
            idmodulo
            for idarea, idmodulo, descripcion_modulo, icono_modulo, version_modulo in self._modules
        ]:
            self._modules.append([area, modulo, descripcion_modulo, datos_icono, version])
            self.process_files(root_folder, modulo)

    def process_files(self, root_folder: str, id_module: str) -> None:
        """Process folder files."""

        for root, subdirs, files in os.walk(root_folder):
            root_dirs_list = pathlib.Path(root)

            for file_name in files:
                if "test" in root_dirs_list.parts or file_name.startswith("test_"):
                    LOGGER.info("FLFILES_DIR: ignoring test %s" % os.path.join(root, file_name))
                    continue

                if file_name.endswith((".pyc")):
                    continue

                if file_name not in [nombre for idmodule, nombre, sha, contenido in self._files]:
                    try:
                        fichero = open(
                            os.path.join(root, file_name),
                            "r",
                            encoding="UTF-8"
                            if file_name.endswith((".ts", ".py"))
                            else "ISO-8859-15",
                        )
                        # print("Guardando ...", os.path.join(root, file_name))
                        data = fichero.read()
                        byte_data = data.encode()
                        sha_ = hashlib.new("sha1", byte_data)
                        string_sha = str(sha_.hexdigest()).upper()
                        self._files.append([id_module, file_name, string_sha, data])
                    except Exception as error:
                        LOGGER.error("Error processing %s:%s", file_name, str(error))
                        return
                else:
                    LOGGER.warning("FLFILES_DIR: file %s already loaded, ignoring..." % file_name)

            for sub_dir in subdirs:
                self.process_files(os.path.join(root_folder, sub_dir), id_module)

            break  # despues de los subdirs salimos...



"""Test_qbytearray module."""

import unittest

from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.core import decorators
from PyQt6 import QtCore, QtWidgets  # type: ignore[import]


class TestQHttp(unittest.TestCase):
    """TestQhttp class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    _state: int
    _buffer: QtCore.QBuffer

    def test_request(self) -> None:
        from pineboolib.q3widgets import qhttp, qbytearray

        host_ = "https://app.slack.com"
        page_ = "client/T0B8PHNUD/C0BA5RGTA"
        http_ = qhttp.QHttp()

        # http_.stateChanged.connect(changeState)
        # http_.dataSendProgress.connect(progressSend)
        http_.dataReadProgress.connect(progressRead)
        http_.requestStarted.connect(startRequest)
        http_.requestFinished.connect(finishRequest)
        http_.done.connect(allDone)

        ba_ = qbytearray.QByteArray()
        self._buffer = QtCore.QBuffer(ba_)

        header_ = qhttp.QHttpRequestHeader("GET", page_)
        header_.setValue("Host", host_)

        http_.setHost(host_)
        http_.request(header_, QtCore.QByteArray(), self._buffer)
        while http_._reply.isRunning():
            QtWidgets.QApplication.processEvents()

        self.assertTrue(len(self._buffer.data()) > 0 or True)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


# @decorators.pyqt_slot(int, int)
# def progressSend(done_: int, total_: int) -> None:
#    """Send progress slot."""

#    print("Enviando", done_, total_)


@decorators.pyqt_slot(int, int)
def progressRead(done_: int, total_: int) -> None:  # pylint: disable=invalid-name
    """Send progress slot."""

    print("Recibiendo", done_, total_)


@decorators.pyqt_slot(int)
def startRequest(id_: int) -> None:  # pylint: disable=invalid-name
    """Send progress slot."""

    print("Iniciando petición", id_)


@decorators.pyqt_slot(int, bool)
def finishRequest(id_: int, error_: bool) -> None:  # pylint: disable=invalid-name
    """Send progress slot."""
    resultado = "OK" if not error_ else "ERROR"

    print("Finalizando petición", id_, resultado)


@decorators.pyqt_slot(bool)
def allDone(error_: bool) -> None:  # pylint: disable=invalid-name
    """Send progress slot."""
    resultado = "OK" if not error_ else "ERROR"

    print("Comunicación realizada", resultado)


# @decorators.pyqt_slot(int)
# def changeState(state_: int) -> None:
#    """State Changed slot."""
#    print("changeState to", state_)



"""Flsettings module."""
# -*- coding: utf-8 -*-
from pineboolib.core import settings
from pineboolib.core.utils import utils_base


from typing import List, Union, Any
from typing import SupportsFloat


class FLSettings(object):
    """FLSettings class."""

    _settings = settings.SETTINGS

    def readListEntry(self, key: str) -> List[str]:
        """Return a value list."""
        ret_ = self._settings.value(key, [])
        if not isinstance(ret_, list):
            ret_ = ret_.split(",")

        return ret_

    def readEntry(self, _key: str, _def: Any = None) -> Any:
        """Return a value."""

        return self._settings.value(_key, _def)

    def readNumEntry(self, key: str, _def: int = 0) -> int:
        """Return a int value."""

        return int(self._settings.value(key, _def))

    def readDoubleEntry(self, key: str, _def: Union[bytes, str, SupportsFloat] = 0.00) -> float:
        """Return a float value."""

        return float(self._settings.value(key, _def))

    def readBoolEntry(self, key: str, _def: bool = False) -> bool:
        """Return a bool value."""

        return utils_base.text2bool(str(self._settings.value(key, _def)))

    def writeEntry(self, key: str, value: Any) -> None:
        """Set a value."""

        self._settings.setValue(key, value)  # type: ignore [has-type]

    def writeEntryList(self, key: str, value: List[str]) -> None:
        """Set a value list."""
        # FIXME: This function flattens the array when saving in some cases. Should always save an array.

        self._settings.setValue(key, ",".join(value))  # type: ignore [has-type]



"""Flqpsql module."""

from pineboolib.application.metadata import pntablemetadata
from pineboolib.application.parsers.parser_mtd import pnmtdparser
from pineboolib import logging

from pineboolib.fllegacy import flutil
from pineboolib.interfaces import isqldriver

from typing import Optional, Union, List, Dict, Any, TYPE_CHECKING

LOGGER = logging.get_logger(__name__)

if TYPE_CHECKING:
    from pineboolib.application.metadata import pnfieldmetadata


class FLQPSQL(isqldriver.ISqlDriver):
    """FLQPSQL class."""

    def __init__(self):
        """Inicialize."""
        super().__init__()
        self.version_ = "0.9"
        self.name_ = "FLQPSQL"
        self.error_list = []
        self.alias_ = "PostgreSQL (PSYCOPG2)"
        self.default_port = 5432
        self._true = True
        self._false = False
        self._like_true = "'t'"
        self._like_false = "'f'"
        self._database_not_found_keywords = ["does not exist", "no existe"]
        self._sqlalchemy_name = "postgresql"
        self._type_array: Dict[int, str] = {
            16: "bool",
            20: "uint",
            23: "uint",
            25: "stringlist",
            701: "double",
            1082: "date",
            1043: "string",
            1184: "timestamp",
            114: "json",
        }

    def getAlternativeConn(self, name: str, host: str, port: int, usern: str, passw_: str) -> Any:
        """Return connection."""
        # import psycopg2

        conn_ = self.getConn("postgres", host, port, usern, passw_)
        # if conn_ is not None:
        #    conn_.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)

        return conn_

    def nextSerialVal(self, table_name: str, field_name: str) -> int:
        """Return next serial value."""

        if self.is_open():
            seq_ = "%s_%s_seq" % (table_name, field_name)
            result_ = 0
            qry = self.execute_query("SELECT NEXTVAL('%s')" % seq_)
            if qry is not None:
                result_ = qry.fetchone()[0]  # type: ignore [index] # noqa: F821
            else:
                self.execute_query("CREATE SEQUENCE %s" % seq_)

        return result_

    def setType(self, type_: str, leng: int = 0) -> str:
        """Return type definition."""
        type_ = type_.lower()

        type_array = {
            "int": "INT2",
            "uint": "INT4",
            "bool": "BOOLEAN",
            "unlock": "BOOLEAN",
            "double": "FLOAT8",
            "time": "TIME",
            "date": "DATE",
            "pixmap": "TEXT",
            "stringlist": "TEXT",
            "string": "VARCHAR",
            "bytearray": "BYTEA",
            "timestamp": "TIMESTAMPTZ",
            "json": "JSON",
        }

        res_ = type_array[type_] if type_ in type_array.keys() else ""

        if not res_:
            LOGGER.warning("seType: unknown type %s", type_)
            leng = 0

        return "%s(%s)" % (res_, leng) if leng else res_

    def sqlCreateTable(
        self, tmd: "pntablemetadata.PNTableMetaData", create_index: bool = True
    ) -> Optional[str]:
        """Return a create table query."""

        if tmd.isQuery():
            return self.sqlCreateView(tmd)

        util = flutil.FLUtil()
        primary_key = ""

        field_list = tmd.fieldList()

        unlocks = 0
        sql_fields: List[str] = []
        for field in field_list:
            sql_field = field.name()
            type_ = field.type()
            if type_ == "serial":
                seq = "%s_%s_seq" % (tmd.name(), field.name())
                if self.is_open() and create_index:
                    cursor = self.execute_query(
                        "SELECT relname FROM pg_class WHERE relname='%s'" % seq
                    )

                    res_ = cursor.fetchone() if cursor else None
                    if not res_:
                        try:
                            self.execute_query("CREATE SEQUENCE %s" % seq)
                        except Exception as error:
                            LOGGER.error("%s::sqlCreateTable:%s", __name__, str(error))

                sql_field += " INT4 DEFAULT NEXTVAL('%s')" % seq
            else:
                if type_ == "unlock":
                    unlocks += 1

                    if unlocks > 1:
                        LOGGER.warning("FLManager : No se ha podido crear la tabla %s ", tmd.name())
                        LOGGER.warning(
                            "FLManager : Hay mas de un campo tipo unlock. Solo puede haber uno."
                        )
                        return None

                sql_field += " %s" % self.setType(type_, field.length())

            if field.isPrimaryKey():
                if not primary_key:
                    sql_field += " PRIMARY KEY"
                    primary_key = field.name()
                else:
                    LOGGER.warning(
                        util.translate(
                            "application",
                            "FLManager : Tabla-> %s ." % tmd.name()
                            + "Se ha intentado poner una segunda clave primaria para el campo %s ,pero el campo %s ya es clave primaria."
                            % (primary_key, field.name())
                            + "Sólo puede existir una clave primaria en FLTableMetaData, use FLCompoundKey para crear claves compuestas.",
                        )
                    )
                    raise Exception(
                        "A primary key (%s) has been defined before the field %s.%s -> %s"
                        % (primary_key, tmd.name(), field.name(), sql_fields)
                    )
            else:
                sql_field += " UNIQUE" if field.isUnique() else ""
                sql_field += " NULL" if field.allowNull() else " NOT NULL"

            sql_fields.append(sql_field)

        return "CREATE TABLE %s (%s)" % (tmd.name(), ",".join(sql_fields))

    def recordInfo2(self, tablename: str) -> Dict[str, List[Any]]:
        """Return info from a database table."""
        info = {}
        sql = (
            "select pg_attribute.attname, pg_attribute.atttypid, pg_attribute.attnotnull, pg_attribute.attlen, pg_attribute.atttypmod, "
            "pg_get_expr(pg_attrdef.adbin, pg_attrdef.adrelid) from pg_class, pg_attribute "
            "left join pg_attrdef on (pg_attrdef.adrelid = pg_attribute.attrelid and pg_attrdef.adnum = pg_attribute.attnum)"
            " where lower(pg_class.relname) = '%s' and pg_attribute.attnum > 0 and pg_attribute.attrelid = pg_class.oid "
            "and pg_attribute.attisdropped = false order by pg_attribute.attnum" % tablename.lower()
        )
        cursor = self.execute_query(sql)
        # res = cursor.fetchall() if cursor else []
        # for columns in res:
        #    field_size = columns[3]
        #    field_precision = columns[4]
        #    field_name = columns[0]
        #    field_type = columns[1]
        #    field_allow_null = columns[2]
        #    field_default_value = columns[5]

        for (
            field_name,
            field_type,
            field_allow_null,
            field_size,
            field_precision,
            field_default_value,
        ) in list(cursor.fetchall() if cursor else []):
            if isinstance(field_default_value, str) and field_default_value:
                field_default_value = (
                    field_default_value[0 : field_default_value.find("::character varying")]
                    if field_default_value.find("::character varying") > -1
                    else field_default_value
                )

            if (
                field_size == -1  # type: ignore [comparison-overlap]
                and field_precision > -1  # type: ignore [operator]
            ):
                field_size = field_precision - 4  # type: ignore [operator, assignment]
                field_precision = -1  # type: ignore [assignment]

            if field_size < 0:  # type: ignore [operator]
                field_size = 0  # type: ignore [assignment]

            if field_precision < 0:  # type: ignore [operator]
                field_precision = 0  # type: ignore [assignment]

            field_default_value = (
                field_default_value[1 : len(field_default_value) - 2]
                if field_default_value and field_default_value[0] == "'"
                else field_default_value
            )

            info[field_name] = [
                field_name,
                self.decodeSqlType(field_type),
                field_allow_null,
                field_size,
                field_precision,
                None,  # defualt_value
                None,  # is_pk
            ]

        return info

    def decodeSqlType(self, type_: Union[int, str]) -> str:
        """Return the specific field type."""

        return self._type_array[int(type_)] if int(type_) in self._type_array.keys() else str(type_)

    def tables(self, type_name: str = "", table_name: str = "") -> List[str]:
        """Return a tables list specified by type."""
        table_list: List[str] = []
        result_list: List[Any] = []

        if self.is_open():
            where: List[str] = []

            if type_name in ["Tables", ""]:
                where.append(
                    "(( relkind = 'r' ) AND ( relname !~ '^Inv' ) AND ( relname !~ '^pg_' ))"
                )
            if type_name in ["Views", ""]:
                where.append(
                    "(( relkind = 'v' ) AND ( relname !~ '^Inv' ) AND ( relname !~ '^pg_' ))"
                )
            if type_name in ["SystemTables", ""]:
                where.append("(( relkind = 'r' ) AND ( relname like 'pg_%%' ))")

            if where:
                and_name = " AND relname ='%s'" % (table_name) if table_name else ""

                cursor = self.execute_query(
                    "select relname from pg_class where %s%s ORDER BY relname ASC"
                    % (" OR ".join(where), and_name)
                )
                result_list += cursor.fetchall() if cursor else []

            table_list = [item[0] for item in result_list]

        return table_list

    def constraintExists(self, name: str) -> bool:
        """Return if constraint exists specified by name."""

        sql = (
            "SELECT constraint_name FROM information_schema.table_constraints where constraint_name='%s'"
            % name
        )
        cur = self.execute_query(sql)
        result_: Any = cur.fetchone() if cur else []
        return True if result_ else False

    def vacuum(self):
        """Vacuum tables."""
        table_names = self.db_.tables("Tables")
        self._connection.connection.set_isolation_level(0)
        for table_name in table_names:
            if self.db_.connManager().manager().metadata(table_name) is not None:
                self.execute_query("VACUUM ANALYZE %s" % table_name)
        self._connection.connection.set_isolation_level(1)

    # def fix_query(self, query: str) -> str:
    #    """Fix string."""
    #    # ret_ = query.replace(";", "")
    #    return query

    def checkSequences(self) -> None:
        """Check sequences."""
        util = flutil.FLUtil()
        conn_dbaux = self.db_.connManager().dbAux()
        sql = (
            "select relname from pg_class where ( relkind = 'r' ) " + "and ( relname !~ '^Inv' ) "
            "and ( relname !~ '^pg_' ) and ( relname !~ '^sql_' )"
        )
        cur_sequences = conn_dbaux.execute_query(sql)
        data_list = list(cur_sequences.fetchall() if cur_sequences else [])
        util.createProgressDialog(
            util.translate("application", "Comprobando indices"), len(data_list)
        )

        for number, data in enumerate(data_list):
            table_name = data[0]
            util.setLabelText(util.translate("application", "Creando índices para %s" % table_name))
            util.setProgress(number)
            metadata = self.db_.connManager().manager().metadata(table_name)
            if metadata is None:
                pass
            #    LOGGER.error("checkSequences: %s metadata not found!", table_name)

        util.destroyProgressDialog()
        return

    def resolveAlterColumn(
        self,
        table_name: str,
        field_meta: "pnfieldmetadata.PNFieldMetaData",
        db_value: List[Union[int, str]],
    ):
        """Return string alter column."""

        def_value: Optional[str] = field_meta.defaultValue()
        text_default_value = ""
        if not field_meta.allowNull():
            text_default_value = " ,server_default='%s'" % (
                def_value
                if def_value is not None
                else self.formatValue(field_meta.type(), None, False)
            )

        result = "op.alter_column('%s', '%s', %s)" % (
            table_name,
            field_meta.name(),
            "type_=%s, existing_type=%s, postgresql_using='%s', nullable=%s%s"
            % (
                pnmtdparser.generate_field(field_meta, "sa"),
                pnmtdparser.resolve_type(db_value[1], db_value[3], "sa"),  # type: ignore [arg-type]
                "%s::%s" % (field_meta.name(), self.setType(field_meta.type())),
                field_meta.allowNull(),
                text_default_value,
            ),
        )

        return result



"""pdf_digest module."""

from pineboolib.core.utils import logging
from pineboolib.core import fonts, images as core_images
import os
from pyhanko import stamp  # type: ignore[import]
from pyhanko.pdf_utils import text, images  # type: ignore[import]
from pyhanko.pdf_utils.font import opentype  # type: ignore[import]
from pyhanko.sign import signers, fields  # type: ignore[import]
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter  # type: ignore[import]
from pyhanko.sign.fields import SigSeedSubFilter  # type: ignore[import]
from pyhanko_certvalidator import ValidationContext  # type: ignore[import]

from typing import Tuple, Dict, Any, Optional

LOGGER = logging.get_logger(__name__)


class PdfDigest:
    """PdfDigest class."""

    _pass: str
    _certificate: Any
    _cert_path: str
    _unsigned_pdf_path: str
    _signer: Optional["signers.PdfSigner"]
    _writer: Optional["IncrementalPdfFileWriter"]
    _stamp: Dict[str, Any]

    def __init__(self, file_path: str, cert_path: str, pass_: str = "") -> None:
        """Initialize."""
        self._cert_path = cert_path
        self._certificate = None
        self._pass = pass_
        self._unsigned_pdf_path = file_path
        self._writer = None
        self._signer = None
        self._stamp = {}

    def set_password(self, pass_: str = "") -> None:
        """Set password."""

        self._pass = pass_

    def _load_certificate(self) -> bool:
        """Load certificate."""

        if not self._pass:
            LOGGER.warning("Password is empty!")
            return False
        try:
            LOGGER.warning("Loading certificate %s with psw %s" % (self._cert_path, self._pass))
            self._certificate = signers.SimpleSigner.load_pkcs12(
                pfx_file=self._cert_path, passphrase=self._pass.encode()
            )
        except Exception as error:
            LOGGER.warning("Error loading certificate: %s", str(error))
            return False

        return True

    def _load_signature(self) -> bool:  # type: ignore [empty-body]
        """Load signature."""

        pass

    def _load_unsigned_document(self, load_writer: bool = False) -> bool:
        """Load unsigned document."""

        if not os.path.exists(self._unsigned_pdf_path):
            raise Exception("%s not exits!" % self._unsigned_pdf_path)

        return True

    def set_stamp(
        self,
        label_text: str,
        url: str = "",
        coords: Tuple[int, int, int, int] = (200, 50, 400, 110),
        background_image_path: str = os.path.join(
            os.path.dirname(core_images.__file__), "icons", "signed_stamp.png"
        ),
        font: str = os.path.join(
            os.path.dirname(fonts.__file__), "noto_sans", "NotoSans-Regular.ttf"
        ),
    ) -> bool:
        """Generate stamp."""

        self._stamp = {
            "text": label_text,
            "coords": coords,
            "background": background_image_path,
            "font": font,
            "url": url,
        }

        return True

    def signature_value(self) -> str:  # type: ignore [empty-body]
        """Return SisgnatureValue field value."""

        pass

    def sign(self, dest_file_path: str = "") -> bool:
        """Sign file."""
        self._load_certificate()

        if dest_file_path and os.path.exists(dest_file_path):
            LOGGER.warning("deleting %s" % dest_file_path)
            os.remove(dest_file_path)

        with open(self._unsigned_pdf_path, "rb") as inf:
            self._writer = IncrementalPdfFileWriter(inf)
            if self._stamp:
                fields.append_signature_field(
                    self._writer,
                    sig_field_spec=fields.SigFieldSpec("Signature1", box=self._stamp["coords"]),
                )

            signature_meta = signers.PdfSignatureMetadata(
                field_name="Signature1",
                md_algorithm="sha256",
                # Mark the signature as a PAdES signature
                subfilter=SigSeedSubFilter.PADES,
                # We'll also need a validation context
                # to fetch & embed revocation info.
                validation_context=ValidationContext(allow_fetching=True),
                # Embed relevant OCSP responses / CRLs (PAdES-LT)
                embed_validation_info=False,
                # Tell pyHanko to put in an extra DocumentTimeStamp
                # to kick off the PAdES-LTA timestamp chain.
                use_pades_lta=False,  # True para timestamp
            )

            kwargs = {"signature_meta": signature_meta, "signer": self._certificate}

            if self._stamp:
                if not os.path.exists(self._stamp["font"]):
                    LOGGER.warning("font file %s not found!" % self._stamp["font"])

                if self._stamp["url"]:
                    LOGGER.warning("QR_MODE!")
                    kwargs["stamp_style"] = stamp.QRStampStyle(
                        stamp_text="%(label)s.\nSigned by:%(signer)s",
                        text_box_style=text.TextBoxStyle(
                            font=opentype.GlyphAccumulatorFactory(self._stamp["font"])
                        ),
                    )

                else:
                    LOGGER.warning("IMAGE_MODE!")
                    kwargs["stamp_style"] = stamp.TextStampStyle(
                        stamp_text="%(label)s.\nSigned by:%(signer)s",
                        text_box_style=text.TextBoxStyle(
                            font=opentype.GlyphAccumulatorFactory(self._stamp["font"])
                        ),
                        background=images.PdfImage(self._stamp["background"]),
                    )

                self._signer = signers.PdfSigner(**kwargs)

            pdf_signer = signers.PdfSigner(
                signature_meta, signer=self._certificate, stamp_style=kwargs["stamp_style"]
            )
            if dest_file_path:
                with open(dest_file_path, "wb") as outf:
                    dict_params = {
                        "url": self._stamp["url"] if "url" in self._stamp.keys() else "",
                        "label": self._stamp["text"] if "text" in self._stamp.keys() else "",
                    }
                    pdf_signer.sign_pdf(
                        self._writer, output=outf, appearance_text_params=dict_params
                    )

            return True
